# Mostly from: https://github.com/ylytkin/python-lasvm/blob/main/lasvm/lasvm.py

from typing import List, Union, Tuple, Optional

import numpy as np
from tqdm import tqdm

from lasvm.base_kernel_method import BaseKernelMethod
from sail.models.native.base import BaseEstimator

__all__ = [
    'LaSVM'
]


class LaSVM(BaseKernelMethod, BaseEstimator):
    """The LaSVM online learning algorithm for binary classification
    as described here:
    https://leon.bottou.org/papers/bordes-ertekin-weston-bottou-2005

    Note that some positive and negative samples are needed to
    initialize the model.

    Example
    -------
    x = np.array([[ 1.84, -1.7 ],
                  [-0.52,  0.27],
                  [-0.23, -0.26],
                  [-1.42,  0.17],
                  [ 1.  , -1.  ],
                  [ 0.01,  1.71],
                  [-0.53,  1.7 ],
                  [-0.27,  0.06]])

    y = np.array([1, 1, 0, 0, 1, 0, 0, 1])

    pos_samples = x[:2]
    neg_samples = x[2:4]

    lasvm = LaSVM(pos_samples, neg_samples)
    lasvm.fit(x, y)

    lasvm.score(x, y)  # 0.875

    Parameters
    ----------
    pos_samples : numpy array
        positive samples, shape (n_vectors, n_features)
    neg_samples : numpy array
        negative samples, shape (n_vectors, n_features)
    c : float, default 1
        regularization parameter
    kernel : {'rbf', 'linear', 'poly'}, default 'rbf'
        kernel name
    gamma : 'scale' or float, default 'scale'
        rbf and polynomial kernel gamma parameter (ignored for linear
        kernel). If 'scale' is passed, the actual value of gamma is
        calculated on initialization as::
            1 / x.shape[1] / x.var()
    degree : int, default 3
        polynomial kernel degree parameter (ignored for the rest of
        the kernels)
    coef0 : float, default 0
        polynomial kernel coef0 parameter (ignored for the rest of
        the kernels)
    tol : float
        tolerance parameter
    niter : int
        number of iterations on final step

    Attributes
    ----------
    c : float
        regularization parameter
    kernel : {'rbf', 'linear', 'poly'}
        kernel name
    gamma : 'scale' or float
        rbf and polynomial kernel gamma parameter
    gamma_ : float
        calculated gamma parameter
    degree : int, default 3
        polynomial kernel degree parameter
    coef0 : float, default 0
        polynomial kernel coef0 parameter
    tol : float
        tolerance parameter
    niter : int
        number of iterations on final step
    support_vectors : numpy array
        support vectors, shape (n_vectors, n_features)
    alpha : numpy array
        model coefficients, shape (n_vectors,)
    intercept : float
        model intercept
    target : numpy array
        support vector class labels, shape (n_vectors,)
    kernel_mx : numpy array
        pair-wise kernel values of support vectors,
        shape (n_vectors, n_vectors)
    gradient : numpy array
        gradient, shape (n_vectors,)
    a : numpy array
        lower bounds, shape (n_vectors,)
    b : numpy array
        upper bounds, shape (n_vectors,)
    delta : float
        current delta

    Methods
    -------
    partial_fit(x, y)
        partially fit the model on the training data
    finalize()
        finalize the training process after fitting is done
    fit(x, y)
        fit the model on the training data (equivalent to
        partial_fit with subsequent finalize)
    predict(x)
        predict class labels of the test data
    score(x, y)
        calculate model accuracy on the given data

    Properties
    ----------
    coef_ : numpy array
        coefficients of the separating hyperplane in the linear
        kernel case, shape (n_vectors,)
    """

    ERR = 0.00001

    class GradientPairNotFoundError(Exception):
        def __init__(self):
            super().__init__('could not find a maximum gradient pair')

    def __init__(
            self,
            pos_samples: np.ndarray,
            neg_samples: np.ndarray,
            c: float = 1,
            kernel: str = 'rbf',
            gamma: Union[float, str] = 'scale',
            degree: int = 3,
            coef0: float = 0,
            tol: float = 0.0001,
            niter: int = 10000,
    ) -> None:
        super().__init__(kernel=kernel, gamma=gamma, degree=degree, coef0=coef0)

        self.c = c

        self.tol = tol

        self.target = np.empty(shape=(0,))
        self.kernel_mx = np.empty(shape=(0, 0))
        self.gradient = np.empty(shape=(0,))
        self.a = np.empty(shape=(0,))
        self.b = np.empty(shape=(0,))

        self.delta = None
        self.niter = niter

        self._initialize(pos_samples, neg_samples)

    def partial_fit(
            self,
            x: np.ndarray,
            y: np.ndarray,
            verbose: bool = False,
            shuffle: bool = False,
    ) -> 'LaSVM':
        """Partially (incrementally) fit model to the given data.

        Array `x` must be 2-dimensional of shape (n_vectors, n_features).
        Array `y` must be 1-dimensional of shape (n_vectors,) and contain only
        class labels 0 or 1.

        Parameters
        ----------
        x : numpy array
            data, shape (n_vectors, n_features)
        y : numpy array
            class labels, shape (n_vectors,), valued 0 or 1
        shuffle : bool
            shuffle data before fitting
        verbose : bool
            set verbosity

        Returns
        -------
        self
        """

        x = x.copy()
        y = self._prepare_targets(y)

        ids = np.arange(x.shape[0])

        if shuffle:
            np.random.shuffle(ids)

        for i in tqdm(ids, disable=not verbose):
            x0 = x[[i]]
            y0 = y[[i]]

            self._process(x=x0, y=y0)
            self._reprocess()

        return self

    def fit(
            self,
            x: np.ndarray,
            y: np.ndarray,
            verbose: bool = False,
            shuffle: bool = False,
    ) -> 'LaSVM':
        """Fit model to the given data. Equivalent to partial_fit with
        subsequent finalize.

        Array `x` must be 2-dimensional of shape (n_vectors, n_features).
        Array `y` must be 1-dimensional of shape (n_vectors,) and contain only
        class labels 0 or 1.

        Parameters
        ----------
        x : numpy array
            data, shape (n_vectors, n_features)
        y : numpy array
            class labels, shape (n_vectors,), valued 0 or 1
        shuffle : bool
            shuffle data before fitting
        verbose : bool
            set verbosity

        Returns
        -------
        self
        """

        self.partial_fit(x, y, shuffle=shuffle, verbose=verbose)

        if verbose:
            print('Finalizing')

        self.finalize(verbose=verbose)

        return self

    def finalize(self, niter: Optional[int] = None, verbose: bool = False) -> 'LaSVM':
        """Perform the final step of learning (i.e. attempt to get delta
        below tau).

        Parameters
        ----------
        niter : int
            number of finalizing iterations
        verbose : bool
            set verbosity

        Returns
        -------
        self
        """

        niter = niter or self.niter

        tqdm_bar = tqdm(total=niter) if verbose else None
        broke_loop = False

        for _ in range(niter):
            self._reprocess()

            if self.delta <= self.tol:
                broke_loop = True
                break

            if verbose:
                tqdm_bar.update(1)

        if not broke_loop:
            print(f'Warning: delta did not converge below tol in {niter} iterations')

        if verbose:
            tqdm_bar.update(tqdm_bar.total - tqdm_bar.n)
            tqdm_bar.close()

        return self

    def __repr__(self):
        return 'LaSVM()'

    def _initialize(self, pos_samples: np.ndarray, neg_samples: np.ndarray) -> 'LaSVM':
        """Initialize model by adding some positive and negative samples as
        support vectors.

        Parameters
        ----------
        pos_samples : numpy array
            positive samples, shape (n_vectors, n_features)
        neg_samples : numpy array
            negative samples, shape (n_vectors, n_features)

        Returns
        -------
        self
        """

        self._remove_all_support_vectors()

        pos_samples = pos_samples.copy()
        neg_samples = neg_samples.copy()

        if self.gamma == 'scale':
            self.gamma_ = self._scaled_gamma(np.vstack([pos_samples, neg_samples]))

        self.support_vectors = np.empty(shape=(0, pos_samples.shape[1]))

        self._add_support_vectors(pos_samples, y=np.ones(pos_samples.shape[0]))
        self._add_support_vectors(neg_samples, y=- np.ones(neg_samples.shape[0]))

        i, j = self._find_maximum_gradient_pair()
        self.intercept = (self.gradient[i] + self.gradient[j]) / 2
        self.delta = self.gradient[i] - self.gradient[j]

        return self

    def _add_support_vectors(self, x: np.ndarray, y: np.ndarray) -> None:
        """Add support vectors with zero coefficients.

        Parameters
        ----------
        x : numpy array
            data, shape (n_vectors, n_features)
        y : numpy array
            class labels, shape (n_vectors, n_features)
        """

        n_vectors = x.shape[0]

        self.support_vectors = np.vstack([self.support_vectors, x])
        self.alpha = np.append(self.alpha, np.zeros(n_vectors))
        self.target = np.append(self.target, y)

        new_kernel_values = self._kernel(x, self.support_vectors)

        self.kernel_mx = np.vstack([self.kernel_mx, new_kernel_values[:, :-n_vectors]])
        self.kernel_mx = np.hstack([self.kernel_mx, new_kernel_values.T])

        gradient = y - new_kernel_values.dot(self.alpha)
        self.gradient = np.append(self.gradient, gradient)

        a = y * self.c
        print(type(a))
        a[a > 0] = 0
        self.a = np.append(self.a, a)

        b = y * self.c
        b[b < 0] = 0
        self.b = np.append(self.b, b)

    def _remove_support_vectors(self, vector_ids: List[int]) -> None:
        """Remove support vectors with given ids.

        Parameters
        ----------
        vector_ids : list
            ids of vectors to remove
        """

        self.support_vectors = np.delete(self.support_vectors, vector_ids, axis=0)
        self.alpha = np.delete(self.alpha, vector_ids)
        self.kernel_mx = np.delete(self.kernel_mx, vector_ids, axis=0)
        self.kernel_mx = np.delete(self.kernel_mx, vector_ids, axis=1)
        self.target = np.delete(self.target, vector_ids)
        self.gradient = np.delete(self.gradient, vector_ids)
        self.a = np.delete(self.a, vector_ids)
        self.b = np.delete(self.b, vector_ids)

    def _remove_all_support_vectors(self) -> None:
        n_sv = self.support_vectors.shape[0]

        if n_sv == 0:
            return

        to_remove = list(range(n_sv))
        self._remove_support_vectors(to_remove)

    def _is_violating_pair(self, i: int, j: int) -> bool:
        return self.alpha[i] < self.b[i] \
               and self.alpha[j] > self.a[j] \
               and self.gradient[i] - self.gradient[j] > self.tol

    def _find_max_gradient_id(self) -> int:
        """Find id of the vector with conditionally maximal gradient.

        Returns
        -------
        int
            vector id
        """

        mask = self.alpha < self.b
        mask_ids = np.where(mask)[0]
        i = mask_ids[np.argmax(self.gradient[mask])]

        return i

    def _find_min_gradient_id(self) -> int:
        """Find id of the vector with conditionally minimal gradient.

        Returns
        -------
        int
            vector id
        """

        mask = self.alpha > self.a
        mask_ids = np.where(mask)[0]
        j = mask_ids[np.argmin(self.gradient[mask])]

        return j

    def _find_maximum_gradient_pair(self) -> Tuple[int, int]:
        return self._find_max_gradient_id(), self._find_min_gradient_id()

    def _update_parameters(self, i: int, j: int) -> None:
        lambda_ = min(
            (self.gradient[i] - self.gradient[j]) / (
                        self.kernel_mx[i, i] + self.kernel_mx[j, j] - 2 * self.kernel_mx[i, j]),
            self.b[i] - self.alpha[i],
            self.alpha[j] - self.a[j],
        )

        self.alpha[i] = self.alpha[i] + lambda_
        self.alpha[j] = self.alpha[j] - lambda_

        self.gradient = self.gradient - lambda_ * (self.kernel_mx[i] - self.kernel_mx[j])

    def _process(self, x: np.ndarray, y: np.ndarray) -> None:
        """Process an object-target pair.

        Parameters
        ----------
        x : numpy array
            feature vector, shape (1, n_features)
        y : numpy array
            class label, shape (1,)
        """

        if (((self.support_vectors - x) ** 2).sum(axis=1) ** 0.5 < self.ERR).any():
            return

        self._add_support_vectors(x, y)

        if y[0] == 1:
            i = self.support_vectors.shape[0] - 1
            j = self._find_min_gradient_id()

        else:
            j = self.support_vectors.shape[0] - 1
            i = self._find_max_gradient_id()

        if not self._is_violating_pair(i, j):
            return

        self._update_parameters(i=i, j=j)

    def _reprocess(self) -> None:
        """Reprocess.
        """

        i, j = self._find_maximum_gradient_pair()

        if not self._is_violating_pair(i, j):
            return

        self._update_parameters(i=i, j=j)

        i, j = self._find_maximum_gradient_pair()
        to_remove = []

        for k in np.where(np.abs(self.alpha) < self.ERR)[0]:
            if (self.target[k] == -1 and self.gradient[k] >= self.gradient[i]) \
                    or (self.target[k] == 1 and self.gradient[k] <= self.gradient[j]):
                to_remove.append(k)

        self.intercept = (self.gradient[i] + self.gradient[j]) / 2
        self.delta = self.gradient[i] - self.gradient[j]

        self._remove_support_vectors(to_remove)




import torch
import torch.nn as nn
from skorch.regressor import NeuralNetRegressor
from sail.models.torch.base import TorchSerializationMixin


class LSTMModel(nn.Module):
    def __init__(self, ni=6, no=3, nh=10, nlayers=1):
        super(LSTMModel, self).__init__()

        self.ni = ni
        self.no = no
        self.nh = nh
        self.nlayers = nlayers

        self.lstms = nn.ModuleList(
            [nn.LSTMCell(self.ni, self.nh)]
            + [nn.LSTMCell(self.nh, self.nh) for i in range(nlayers - 1)]
        )

        self.out = nn.Linear(self.nh, self.no)
        self.do = nn.Dropout(p=0.2)
        self.actfn = nn.Tanh()
        self.device = torch.device("cpu")
        self.dtype = torch.float

    # description of the whole block
    def forward(self, x, h0=None, train=False):
        hs = x  # initiate hidden state
        if h0 is None:
            h = torch.zeros(hs.shape[0], self.nh)
            c = torch.zeros(hs.shape[0], self.nh)
        else:
            (h, c) = h0

        # LSTM cells
        for i in range(self.nlayers):
            h, c = self.lstms[i](hs, (h, c))
            if train:
                hs = self.do(h)
            else:
                hs = h
        y = self.out(hs)
        return y, (h, c)


class ContextlessMSE(torch.nn.MSELoss):
    def forward(self, y_pred, y_true):
        y, (h, c) = y_pred  # extract prediction and context information
        return super().forward(y, y_true)


class LSTMRegressor(NeuralNetRegressor, TorchSerializationMixin):
    def __init__(self, ni, no, nh, nlayers, module=LSTMModel, **kwargs):
        super(LSTMRegressor, self).__init__(
            module=module,
            module__ni=ni,
            module__no=no,
            module__nh=nh,
            module__nlayers=nlayers,
            **kwargs
        )



import logging
import logzero
import sail


def configure_logger(
    logger_name: str,
    logging_level=sail.get_logging_level(),
    json=False,
):
    if logger_name is None or logger_name == "":
        raise Exception(
            "Logger name cannot be None or empty. Accessing root logger is restricted. Please use routine configure_root_logger() to access the root logger."
        )

    heading = "SAIL" + " " + "(" + logger_name + ")"

    info_log_format = (
        "%(color)s[%(asctime)s:%(msecs)d] - %(levelname)s - " + heading + " - "
        "%(end_color)s%(message)s"
    )
    rest_log_format = (
        "%(color)s[%(asctime)s:%(msecs)d] - %(levelname)s - " + heading + " - "
        "%(module)s:%(funcName)s:%(lineno)d%(end_color)s %(message)s"
    )

    logging_level = logging.getLevelName(logging_level)
    if logging.INFO == logging_level:
        log_format = info_log_format
    else:
        log_format = rest_log_format

    formatter = logzero.LogFormatter(fmt=log_format, datefmt="%Y-%m-%d %H:%M:%S")
    return logzero.setup_logger(
        name=logger_name, level=logging_level, formatter=formatter, json=json
    )


def configure_root_logger(
    logging_level=sail.get_logging_level(),
    json=False,
):
    heading = "SAIL"

    info_log_format = (
        "%(color)s[%(asctime)s:%(msecs)d] - %(levelname)s - "
        + heading
        + "%(end_color)s%(message)s"
    )
    rest_log_format = (
        "%(color)s[%(asctime)s:%(msecs)d] - %(levelname)s - "
        + heading
        + "%(module)s:%(funcName)s:%(lineno)d%(end_color)s %(message)s"
    )

    logging_level = logging.getLevelName(logging_level)
    if logging.INFO == logging_level:
        log_format = info_log_format
    else:
        log_format = rest_log_format

    formatter = logzero.LogFormatter(fmt=log_format, datefmt="%Y-%m-%d %H:%M:%S")
    return logzero.setup_logger(
        name=None, level=logging_level, formatter=formatter, json=json
    )



from abc import ABC, abstractmethod
from river.compat.river_to_sklearn import SKLEARN_INPUT_X_PARAMS, SKLEARN_INPUT_Y_PARAMS

from sklearn.base import BaseEstimator, TransformerMixin


class BaseTransformer(ABC):
    @abstractmethod
    def fit():
        raise NotImplementedError

    @abstractmethod
    def partial_fit():
        raise NotImplementedError

    @abstractmethod
    def transform():
        raise NotImplementedError

    @abstractmethod
    def fit_transform():
        raise NotImplementedError

    @abstractmethod
    def partial_fit_transform():
        raise NotImplementedError


class SAILTransformer(TransformerMixin, BaseEstimator, BaseTransformer):
    def __init__(self) -> None:
        super(SAILTransformer, self).__init__()
        self.validation_params = {"cast_to_ndarray": True}

    def fit(self, X, y=None):
        # Reset the state if already fitted
        for attr in ("instance_", "n_features_in_"):
            self.__dict__.pop(attr, None)

        # Fit with one pass of the dataset
        return self.partial_fit(X, y)

    def partial_fit(self, X, y=None):
        # Check the inputs
        first_call = not hasattr(self, "n_features_in_")
        if y is None:
            X = self._validate_data(
                X,
                **self.validation_params,
                reset=first_call,
                **SKLEARN_INPUT_X_PARAMS,
            )
        else:
            X, y = self._validate_data(
                X,
                y,
                **self.validation_params,
                reset=first_call,
                **SKLEARN_INPUT_X_PARAMS,
                **SKLEARN_INPUT_Y_PARAMS,
            )

        # Fit with one pass of the dataset
        return self._partial_fit(X, y)

    def transform(self, X, y=None, copy=None):
        # Check the inputs
        X = self._validate_data(
            X,
            **self.validation_params,
            reset=False,
            **SKLEARN_INPUT_X_PARAMS,
        )

        # Fit with one pass of the dataset
        return self._transform(X, y, copy)

    def partial_fit_transform(self, X, y=None, **fit_params):
        if y is None:
            # fit method of arity 1 (unsupervised transformation)
            return self.partial_fit(X, **fit_params).transform(X)
        else:
            # fit method of arity 2 (supervised transformation)
            return self.partial_fit(X, y, **fit_params).transform(X)



import river.naive_bayes as naive_bayes
from sail.models.river.base import SailRiverClassifier


__all__ = ["GaussianNB", "MultinomialNB", "ComplementNB", "BernoulliNB"]


class GaussianNB(SailRiverClassifier):
    def __init__(self):
        super(GaussianNB, self).__init__(river_estimator=naive_bayes.GaussianNB())


class MultinomialNB(SailRiverClassifier):
    def __init__(self, alpha=1.0):
        super(MultinomialNB, self).__init__(
            river_estimator=naive_bayes.MultinomialNB(alpha)
        )


class ComplementNB(SailRiverClassifier):
    def __init__(self, alpha=1.0):
        super(ComplementNB, self).__init__(
            river_estimator=naive_bayes.ComplementNB(alpha)
        )


class BernoulliNB(SailRiverClassifier):
    def __init__(self, alpha=1.0, true_threshold=0.0):
        super(BernoulliNB, self).__init__(
            river_estimator=naive_bayes.BernoulliNB(alpha, true_threshold)
        )



""" Base class for all estimators in sail """

# Almost all the code taken from scikit-multiflow
# https://github.com/scikit-multiflow/scikit-multiflow/blob/a7e316d/src/skmultiflow/core/base.py

import copy
import inspect
import re
import warnings
from abc import ABCMeta, abstractmethod
from collections import defaultdict

import sklearn

sklearn.set_config(print_changed_only=False)
import numpy as np
from sklearn.utils._pprint import _EstimatorPrettyPrinter

__version__ = "0.1.0"

_DEFAULT_TAGS = {
    "non_deterministic": False,
    "requires_positive_data": False,
    "X_types": ["2darray"],
    "poor_score": False,
    "no_validation": False,
    "multioutput": False,
    "allow_nan": False,
    "stateless": False,
    "multilabel": False,
    "_skip_test": False,
    "multioutput_only": False,
}


def clone(estimator, safe=True):
    """Constructs a new estimator with the same parameters.
    Clone does a deep copy of the model in an estimator
    without actually copying attached data. It yields a new estimator
    with the same parameters that has not been fit on any data.
    Parameters
    ----------
    estimator : estimator object, or list, tuple or set of objects
        The estimator or group of estimators to be cloned
    safe : boolean, optional
        If safe is false, clone will fall back to a deep copy on objects
        that are not estimators.
    Notes
    -----
    Taken from sklearn for compatibility.
    """
    estimator_type = type(estimator)
    # XXX: not handling dictionaries
    if estimator_type in (list, tuple, set, frozenset):
        return estimator_type([clone(e, safe=safe) for e in estimator])
    elif not hasattr(estimator, "get_params") or isinstance(estimator, type):
        if not safe:
            return copy.deepcopy(estimator)
        else:
            raise TypeError(
                "Cannot clone object '%s' (type %s): "
                "it does not seem to be a valid estimator "
                "as it does not implement a 'get_params' methods."
                % (repr(estimator), type(estimator))
            )
    klass = estimator.__class__
    new_object_params = estimator.get_params(deep=False)
    for name, param in new_object_params.items():
        new_object_params[name] = clone(param, safe=False)
    new_object = klass(**new_object_params)
    params_set = new_object.get_params(deep=False)

    # quick sanity check of the parameters of the clone
    for name in new_object_params:
        param1 = new_object_params[name]
        param2 = params_set[name]
        if param1 is not param2:
            raise RuntimeError(
                "Cannot clone object %s, as the constructor "
                "either does not set or modifies parameter %s" % (estimator, name)
            )
    return new_object


def _pprint(params, offset=0, printer=repr):
    """Pretty print the dictionary 'params'
    Parameters
    ----------
    params : dict
        The dictionary to pretty print
    offset : int
        The offset in characters to add at the begin of each line.
    printer : callable
        The function to convert entries to strings, typically
        the builtin str or repr
    Notes
    -----
    Taken from sklearn for compatibility.
    """
    # Do a multi-line justified repr:
    options = np.get_printoptions()
    np.set_printoptions(precision=5, threshold=64, edgeitems=2)
    params_list = list()
    this_line_length = offset
    line_sep = ",\n" + (1 + offset // 2) * " "
    for i, (k, v) in enumerate(sorted(params.items())):
        if type(v) is float:
            # use str for representing floating point numbers
            # this way we get consistent representation across
            # architectures and versions.
            this_repr = "%s=%s" % (k, str(v))
        else:
            # use repr of the rest
            this_repr = "%s=%s" % (k, printer(v))
        if len(this_repr) > 500:
            this_repr = this_repr[:300] + "..." + this_repr[-100:]
        if i > 0:
            if this_line_length + len(this_repr) >= 75 or "\n" in this_repr:
                params_list.append(line_sep)
                this_line_length = len(line_sep)
            else:
                params_list.append(", ")
                this_line_length += 2
        params_list.append(this_repr)
        this_line_length += len(this_repr)

    np.set_printoptions(**options)
    lines = "".join(params_list)
    # Strip trailing space to avoid nightmare in doctests
    lines = "\n".join(line.rstrip(" ") for line in lines.split("\n"))
    return lines


def _update_if_consistent(dict1, dict2):
    common_keys = set(dict1.keys()).intersection(dict2.keys())
    for key in common_keys:
        if dict1[key] != dict2[key]:
            raise TypeError(
                "Inconsistent values for tag {}: {} != {}".format(
                    key, dict1[key], dict2[key]
                )
            )
    dict1.update(dict2)
    return dict1


class BaseEstimator:
    """Base Estimator class for compatibility with scikit-learn.
    Notes
    -----
    * All estimators should specify all the parameters that can be set
      at the class level in their ``__init__`` as explicit keyword
      arguments (no ``*args`` or ``**kwargs``).
    * Taken from sklearn for compatibility.
    """

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "scikit-multiflow estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    " %s with constructor %s doesn't "
                    " follow this convention." % (cls, init_signature)
                )
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=True):
        """Get parameters for this estimator.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.
        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.
        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition("__")
            if key not in valid_params:
                raise ValueError(
                    "Invalid parameter %s for estimator %s. "
                    "Check the list of available parameters "
                    "with `estimator.get_params().keys()`." % (key, self)
                )

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def __repr__(self, N_CHAR_MAX=700):
        # N_CHAR_MAX is the (approximate) maximum number of non-blank
        # characters to render. We pass it as an optional parameter to ease
        # the tests.

        N_MAX_ELEMENTS_TO_SHOW = 30  # number of elements to show in sequences

        # use ellipsis for sequences with a lot of elements
        pp = _EstimatorPrettyPrinter(
            compact=True,
            indent=1,
            indent_at_name=True,
            n_max_elements_to_show=N_MAX_ELEMENTS_TO_SHOW,
        )

        repr_ = pp.pformat(self)

        # Use bruteforce ellipsis when there are a lot of non-blank characters
        n_nonblank = len("".join(repr_.split()))
        if n_nonblank > N_CHAR_MAX:
            lim = N_CHAR_MAX // 2  # apprx number of chars to keep on both ends
            regex = r"^(\s*\S){%d}" % lim
            # The regex '^(\s*\S){%d}' % n
            # matches from the start of the string until the nth non-blank
            # character:
            # - ^ matches the start of string
            # - (pattern){n} matches n repetitions of pattern
            # - \s*\S matches a non-blank char following zero or more blanks
            left_lim = re.match(regex, repr_).end()
            right_lim = re.match(regex, repr_[::-1]).end()

            if "\n" in repr_[left_lim:-right_lim]:
                # The left side and right side aren't on the same line.
                # To avoid weird cuts, e.g.:
                # categoric...ore',
                # we need to start the right side with an appropriate newline
                # character so that it renders properly as:
                # categoric...
                # handle_unknown='ignore',
                # so we add [^\n]*\n which matches until the next \n
                regex += r"[^\n]*\n"
                right_lim = re.match(regex, repr_[::-1]).end()

            ellipsis = "..."
            if left_lim + len(ellipsis) < len(repr_) - right_lim:
                # Only add ellipsis if it results in a shorter repr
                repr_ = repr_[:left_lim] + "..." + repr_[-right_lim:]

        return repr_

    def __getstate__(self):
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__.copy()
        return state

    def __setstate__(self, state):
        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(state)

    def _get_tags(self):
        collected_tags = {}
        for base_class in inspect.getmro(self.__class__):
            if hasattr(base_class, "_more_tags") and base_class != self.__class__:
                more_tags = base_class._more_tags(self)
                collected_tags = _update_if_consistent(collected_tags, more_tags)
        if hasattr(self, "_more_tags"):
            more_tags = self._more_tags()
            collected_tags = _update_if_consistent(collected_tags, more_tags)
        tags = _DEFAULT_TAGS.copy()
        tags.update(collected_tags)
        return tags


class BaseSailObject(BaseEstimator):
    """Base class for most objects in sail
    Notes
    -----
    This class provides additional functionality not available in the base estimator
    from scikit-learn
    """

    def reset(self):
        """Resets the estimator to its initial state.
        Returns
        -------
        self
        """
        # non-optimized default implementation; override if a better
        # method is possible for a given object
        command = "".join([line.strip() for line in self.__repr__().split()])
        command = command.replace(str(self.__class__.__name__), "self.__init__")
        exec(command)

    def get_info(self):
        """Collects and returns the information about the configuration of the estimator
        Returns
        -------
        string
            Configuration of the estimator.
        """
        return self.__repr__()


class ClassifierMixin(metaclass=ABCMeta):
    """Mixin class for all classifiers in scikit-multiflow."""

    _estimator_type = "classifier"

    def fit(self, X, y, classes=None, sample_weight=None):
        """Fit the model.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.
        y: numpy.ndarray of shape (n_samples, n_targets)
            An array-like with the class labels of all samples in X.
        classes: numpy.ndarray, optional (default=None)
            Contains all possible/known class labels. Usage varies depending
            on the learning method.
        sample_weight: numpy.ndarray, optional (default=None)
            Samples weight. If not provided, uniform weights are assumed.
            Usage varies depending on the learning method.
        Returns
        -------
        self
        """
        # non-optimized default implementation; override if a better
        # method is possible for a given classifier
        self.partial_fit(X, y, classes=classes, sample_weight=sample_weight)

        return self

    @abstractmethod
    def partial_fit(self, X, y, classes=None, sample_weight=None):
        """Partially (incrementally) fit the model.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.
        y: numpy.ndarray of shape (n_samples)
            An array-like with the class labels of all samples in X.
        classes: numpy.ndarray, optional (default=None)
            Array with all possible/known class labels. Usage varies depending
            on the learning method.
        sample_weight: numpy.ndarray of shape (n_samples), optional (default=None)
            Samples weight. If not provided, uniform weights are assumed.
            Usage varies depending on the learning method.
        Returns
        -------
        self
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, X):
        """Predict classes for the passed data.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of data samples to predict the class labels for.
        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_proba(self, X):
        """Estimates the probability of each sample in X belonging to each of the class-labels.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The matrix of samples one wants to predict the class probabilities for.
        Returns
        -------
        A numpy.ndarray of shape (n_samples, n_labels), in which each outer entry is associated
        with the X entry of the same index. And where the list in index [i] contains
        len(self.target_values) elements, each of which represents the probability that
        the i-th sample of X belongs to a certain class-label.
        """
        raise NotImplementedError

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.
        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.
        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.
        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True labels for X.
        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.
        Returns
        -------
        score : float
            Mean accuracy of self.predict(X) wrt. y.
        """
        from sklearn.metrics import accuracy_score

        return accuracy_score(y, self.predict(X), sample_weight=sample_weight)


class RegressorMixin(metaclass=ABCMeta):
    """Mixin class for all regression estimators in scikit-multiflow."""

    _estimator_type = "regressor"

    def fit(self, X, y, sample_weight=None):
        """Fit the model.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.
        y: numpy.ndarray of shape (n_samples, n_targets)
            An array-like with the target values of all samples in X.
        sample_weight: numpy.ndarray, optional (default=None)
            Samples weight. If not provided, uniform weights are assumed. Usage varies
            depending on the learning method.
        Returns
        -------
        self
        """
        # non-optimized default implementation; override if a better
        # method is possible for a given regressor
        self.partial_fit(X, y, sample_weight=sample_weight)

        return self

    @abstractmethod
    def partial_fit(self, X, y, sample_weight=None):
        """Partially (incrementally) fit the model.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The features to train the model.
        y: numpy.ndarray of shape (n_samples)
            An array-like with the target values of all samples in X.
        sample_weight: numpy.ndarray of shape (n_samples), optional (default=None)
            Samples weight. If not provided, uniform weights are assumed. Usage varies
            depending on the learning method.
        Returns
        -------
        self
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, X):
        """Predict target values for the passed data.
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The set of data samples to predict the target values for.
        Returns
        -------
        A numpy.ndarray with all the predictions for the samples in X.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_proba(self, X):
        """Estimates the probability for probabilistic/bayesian regressors
        Parameters
        ----------
        X : numpy.ndarray of shape (n_samples, n_features)
            The matrix of samples one wants to predict the probabilities for.
        Returns
        -------
        numpy.ndarray
        """
        raise NotImplementedError

    def score(self, X, y, sample_weight=None):
        """Returns the coefficient of determination R^2 of the prediction.
        The coefficient R^2 is defined as (1 - u/v), where u is the residual
        sum of squares ((y_true - y_pred) ** 2).sum() and v is the total
        sum of squares ((y_true - y_true.mean()) ** 2).sum().
        The best possible score is 1.0 and it can be negative (because the
        model can be arbitrarily worse). A constant model that always
        predicts the expected value of y, disregarding the input features,
        would get a R^2 score of 0.0.
        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples. For some estimators this may be a
            precomputed kernel matrix instead, shape = (n_samples,
            n_samples_fitted], where n_samples_fitted is the number of
            samples used in the fitting for the estimator.
        y : array-like, shape = (n_samples) or (n_samples, n_outputs)
            True values for X.
        sample_weight : array-like, shape = [n_samples], optional
            Sample weights.
        Returns
        -------
        score : float
            R^2 of self.predict(X) wrt. y.
        Notes
        -----
        The R2 score used when calling ``score`` on a regressor will use
        ``multioutput='uniform_average'`` from version 0.23 to keep consistent
        with `metrics.r2_score`. This will influence the ``score`` method of
        all the multioutput regressors (except for
        `multioutput.MultiOutputRegressor`). To specify the default value
        manually and avoid the warning, please either call `metrics.r2_score`
        directly or make a custom scorer with `metrics.make_scorer` (the
        built-in scorer ``'r2'`` uses ``multioutput='uniform_average'``).
        """

        from sklearn.metrics import r2_score
        from sklearn.metrics._regression import _check_reg_targets

        y_pred = self.predict(X)
        # XXX: Remove the check in 0.23
        y_type, _, _, _ = _check_reg_targets(y, y_pred, None)
        if y_type == "continuous-multioutput":
            warnings.warn(
                "The default value of multioutput (not exposed in "
                "score method) will change from 'variance_weighted' "
                "to 'uniform_average' in 0.23 to keep consistent "
                "with 'metrics.r2_score'. To specify the default "
                "value manually and avoid the warning, please "
                "either call 'metrics.r2_score' directly or make a "
                "custom scorer with 'metrics.make_scorer' (the "
                "built-in scorer 'r2' uses "
                "multioutput='uniform_average').",
                FutureWarning,
            )
        return r2_score(
            y, y_pred, sample_weight=sample_weight, multioutput="variance_weighted"
        )


class MetaEstimatorMixin(object):
    """Mixin class for all meta estimators in scikit-multiflow."""

    _required_parameters = ["estimator"]


class MultiOutputMixin(object):
    """Mixin to mark estimators that support multioutput."""

    def _more_tags(self):
        return {"multioutput": True}


def is_classifier(estimator):
    """Returns True if the given estimator is (probably) a classifier.
    Parameters
    ----------
    estimator : object
        Estimator object to test.
    Returns
    -------
    out : bool
        True if estimator is a classifier and False otherwise.
    """
    return getattr(estimator, "_estimator_type", None) == "classifier"


def is_regressor(estimator):
    """Returns True if the given estimator is (probably) a regressor.
    Parameters
    ----------
    estimator : object
        Estimator object to test.
    Returns
    -------
    out : bool
        True if estimator is a regressor and False otherwise.
    """
    return getattr(estimator, "_estimator_type", None) == "regressor"



import importlib
import os
import random
import time
import warnings
from operator import *

import numpy as np
from ray import tune
from ray.tune.search import BasicVariantGenerator, ConcurrencyLimiter, Searcher
from ray.tune.search.bohb import TuneBOHB
from ray.tune.search.hyperopt import HyperOptSearch
from ray.tune.search.optuna import OptunaSearch
from ray.tune.search.skopt import SkOptSearch
from ray.tune.sklearn import TuneGridSearchCV, TuneSearchCV
from ray.tune.stopper import CombinedStopper
from sklearn.base import clone
from sklearn.model_selection import ParameterGrid
from tune_sklearn._trainable import _PipelineTrainable, _Trainable
from tune_sklearn.utils import (
    MaximumIterationStopper,
    check_is_pipeline,
    resolve_logger_callbacks,
)

os.environ["TUNE_DISABLE_AUTO_CALLBACK_LOGGERS"] = "1"
os.environ["RAY_DEDUP_LOGS"] = "0"

from sail.models.auto_ml.searcher import SailListSearcher, SailRandomListSearcher


def extract_top_best_configurations(
    num_of_configurations, metric_name, trial_results, params_grid
):
    best_trials = sorted(
        trial_results.items(),
        reverse=True,
        key=lambda trial: getitem(trial[1], metric_name),
    )[0:num_of_configurations]

    eligible_keys = []
    if isinstance(params_grid, list):
        for params in params_grid:
            for param, _ in params.items():
                eligible_keys.append(param)
    else:
        for param, _ in params_grid.items():
            eligible_keys.append(param)

    best_configs = []
    for trial, trial_results in best_trials:
        configs = trial_results["config"]
        configs = {param: configs[param] for param in eligible_keys if param in configs}
        best_configs.append(configs)

    return best_configs


class SAILTuneGridSearchCV(TuneGridSearchCV):
    default_search_params = {
        "verbose": 1,
        "max_iters": 1,
        "early_stopping": False,
        "mode": "max",
        "scoring": "accuracy",
        "pipeline_auto_early_stop": False,
        "keep_best_configurations": 1,
    }
    defined_loggers = ["csv", "json"]

    def __init__(
        self,
        num_cpus_per_trial=1,
        num_gpus_per_trial=0,
        keep_best_configurations=1,
        cluster_address=None,
        storage_path=None,
        namespace=None,
        runtime_env=None,
        **kwargs,
    ):
        super(SAILTuneGridSearchCV, self).__init__(**kwargs)
        self.num_cpus_per_trial = num_cpus_per_trial
        self.num_gpus_per_trial = num_gpus_per_trial
        self.keep_best_configurations = keep_best_configurations
        self.cluster_address = cluster_address
        self.storage_path = storage_path
        self.namespace = namespace
        self.runtime_env = runtime_env

    def fit(
        self, X, y=None, warm_start=False, groups=None, tune_params=None, **fit_params
    ):
        self.warm_start = warm_start
        return super().fit(X, y, groups, tune_params, **fit_params)

    def _list_grid_num_samples(self, warm_start):
        """Calculate the num_samples for `tune.run`.

        This is used when a list of dictionaries is passed in
        for the `param_grid`
        """
        num_samples = 0
        if (
            warm_start
            and hasattr(self, "_best_configurations")
            and self._best_configurations
        ):
            num_samples = len(self._best_configurations)
        return num_samples + len(list(ParameterGrid(self.param_grid)))

    def _num_samples(self, list_searcher=False):
        """Calculate the num_samples for `tune.run`.

        This is used when a dictionaries is passed in
        for the `param_grid`
        """
        num_samples = 0
        if hasattr(self, "_best_configurations") and self._best_configurations:
            num_samples = len(self._best_configurations)
        return num_samples + 1

    def _tune_run(
        self, X, y, config, resources_per_trial, tune_params=None, fit_params=None
    ):
        """Wrapper to call ``tune.run``. Multiple estimators are generated when
        early stopping is possible, whereas a single estimator is
        generated when  early stopping is not possible.

        Args:
            X (:obj:`array-like` (shape = [n_samples, n_features])):
                Training vector, where n_samples is the number of samples and
                n_features is the number of features.
            y (:obj:`array-like`): Shape of array expected to be [n_samples]
                or [n_samples, n_output]). Target relative to X for
                classification or regression; None for unsupervised learning.
            config (dict): Configurations such as hyperparameters to run
                ``tune.run`` on.
            resources_per_trial (dict): Resources to use per trial within Ray.
                Accepted keys are `cpu`, `gpu` and custom resources, and values
                are integers specifying the number of each resource to use.
            tune_params (dict): User defined parameters passed to
                ``tune.run``. Parameters inside `tune_params` override
                preset parameters.
            fit_params (dict): Parameters passed to the ``fit`` method
                of the estimator.

        Returns:
            analysis (`ExperimentAnalysis`): Object returned by
                `tune.run`.

        """
        trainable = _Trainable
        if (
            self.pipeline_auto_early_stop
            and check_is_pipeline(self.estimator)
            and self.early_stopping_
        ):
            trainable = _PipelineTrainable

        if self.early_stopping_ is not None:
            estimator_list = [clone(self.estimator) for _ in range(self.n_splits)]
        else:
            estimator_list = [clone(self.estimator)]

        stopper = MaximumIterationStopper(max_iter=self.max_iters)
        if self.stopper:
            stopper = CombinedStopper(stopper, self.stopper)

        run_args = dict(
            scheduler=self.early_stopping_,
            reuse_actors=True,
            verbose=self.verbose,
            stop=stopper,
            config=config,
            fail_fast="raise",
            storage_path=self.storage_path,
            resources_per_trial={
                "cpu": self.num_cpus_per_trial,
                "gpu": self.num_gpus_per_trial,
            },
            trial_dirname_creator=lambda trial: f"Trail_{trial.trial_id}",
            name="SAILAutoML_Experiment" + "_" + time.strftime("%d-%m-%Y_%H:%M:%S"),
            callbacks=resolve_logger_callbacks(self.loggers, self.defined_loggers),
            time_budget_s=self.time_budget_s,
            metric=self._metric_name,
            mode=self.mode,
        )

        if self.warm_start and hasattr(self, "_best_configurations"):
            best_configurations = self._best_configurations
        else:
            best_configurations = None

        if isinstance(self.param_grid, list):
            run_args.update(
                dict(
                    search_alg=SailListSearcher(
                        self.param_grid, points_to_evaluate=best_configurations
                    ),
                    num_samples=self._list_grid_num_samples(self.warm_start),
                )
            )
        else:
            max_concurrent_trials = tune_params.get("max_concurrent_trials", 0)
            run_args.update(
                dict(
                    search_alg=BasicVariantGenerator(
                        max_concurrent=max_concurrent_trials,
                        points_to_evaluate=best_configurations,
                    ),
                    num_samples=self._num_samples(),
                )
            )

        run_args = self._override_run_args_with_tune_params(run_args, tune_params)
        # Currently, there is a bug where a tqdm logger instance is left unhandled during the Ray Tune. Also, it is an overkill to log Pipeline Training for Distributed Cross Validation. Hence, turning off the verbosity for `SAILPipeline`. This does not affect the Ray Tune logs which can be set via verbose field of tune.run.
        for estimator in estimator_list:
            estimator.verbosity_level = 0

        trainable = tune.with_parameters(
            trainable,
            X=X,
            y=y,
            estimator_list=estimator_list,
            fit_params=fit_params,
        )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="fail_fast='raise' " "detected.")
            analysis = tune.run(
                trainable,
                **run_args,
            )

        if self.keep_best_configurations > 0:
            self._best_configurations = extract_top_best_configurations(
                self.keep_best_configurations,
                self._metric_name,
                analysis.results,
                self.param_grid,
            )

        return analysis


class SAILTuneSearchCV(TuneSearchCV):
    default_search_params = {
        "verbose": 1,
        "scoring": "accuracy",
        "mode": "max",
        "early_stopping": False,
        "n_trials": 5,
        "search_optimization": "random",
        "pipeline_auto_early_stop": False,
        "keep_best_configurations": 1,
    }
    defined_loggers = ["csv", "json"]

    def __init__(
        self,
        num_cpus_per_trial=1,
        num_gpus_per_trial=0,
        keep_best_configurations=1,
        cluster_address=None,
        storage_path=None,
        **kwargs,
    ):
        super(SAILTuneSearchCV, self).__init__(**kwargs)
        self.num_cpus_per_trial = num_cpus_per_trial
        self.num_gpus_per_trial = num_gpus_per_trial
        self.keep_best_configurations = keep_best_configurations
        self.cluster_address = cluster_address
        self.storage_path = storage_path

    def fit(
        self, X, y=None, warm_start=False, groups=None, tune_params=None, **fit_params
    ):
        self.warm_start = warm_start
        return super().fit(X, y, groups, tune_params, **fit_params)

    def _tune_run(
        self, X, y, config, resources_per_trial, tune_params=None, fit_params=None
    ):
        """Wrapper to call ``tune.run``. Multiple estimators are generated when
        early stopping is possible, whereas a single estimator is
        generated when early stopping is not possible.

        Args:
            X (:obj:`array-like` (shape = [n_samples, n_features])):
                Training vector, where n_samples is the number of samples and
                n_features is the number of features.
            y (:obj:`array-like`): Shape of array expected to be [n_samples]
                or [n_samples, n_output]). Target relative to X for
                classification or regression; None for unsupervised learning.
            config (dict): Configurations such as hyperparameters to run
            ``tune.run`` on.
            resources_per_trial (dict): Resources to use per trial within Ray.
                Accepted keys are `cpu`, `gpu` and custom resources, and values
                are integers specifying the number of each resource to use.
            tune_params (dict): User defined parameters passed to
                ``tune.run``. Parameters inside `tune_params` override
                preset parameters.
            fit_params (dict): Parameters passed to the ``fit`` method
                of the estimator.

        Returns:
            analysis (`ExperimentAnalysis`): Object returned by
                `tune.run`.

        """
        if self.seed is not None:
            random.seed(self.seed)
            np.random.seed(self.seed)

        trainable = _Trainable
        if (
            self.pipeline_auto_early_stop
            and check_is_pipeline(self.estimator)
            and self.early_stopping_
        ):
            trainable = _PipelineTrainable

        max_iter = self.max_iters
        if self.early_stopping_ is not None:
            estimator_list = [clone(self.estimator) for _ in range(self.n_splits)]
            if hasattr(self.early_stopping_, "_max_t_attr"):
                # we want to delegate stopping to schedulers which
                # support it, but we want it to stop eventually, just in case
                # the solution is to make the stop condition very big
                max_iter = self.max_iters * 10
        else:
            estimator_list = [clone(self.estimator)]

        stopper = MaximumIterationStopper(max_iter=max_iter)
        if self.stopper:
            stopper = CombinedStopper(stopper, self.stopper)
        run_args = dict(
            scheduler=self.early_stopping_,
            reuse_actors=True,
            verbose=self.verbose,
            stop=stopper,
            num_samples=self.n_trials,
            config=config,
            fail_fast="raise",
            storage_path=self.storage_path,
            resources_per_trial={
                "cpu": self.num_cpus_per_trial,
                "gpu": self.num_gpus_per_trial,
            },
            local_dir=self.local_dir,
            name="SAILAutoML_Experiment" + "_" + time.strftime("%d-%m-%Y_%H:%M:%S"),
            callbacks=resolve_logger_callbacks(self.loggers, self.defined_loggers),
            time_budget_s=self.time_budget_s,
            metric=self._metric_name,
            mode=self.mode,
        )

        if self.warm_start and hasattr(self, "_best_configurations"):
            best_configurations = self._best_configurations
        else:
            best_configurations = None

        if self._search_optimization_lower == "random":
            if isinstance(self.param_distributions, list):
                search_algo = SailRandomListSearcher(
                    self.param_distributions, points_to_evaluate=best_configurations
                )
            else:
                search_algo = BasicVariantGenerator(
                    points_to_evaluate=best_configurations
                )
            run_args["search_alg"] = search_algo
        else:
            search_space = None
            override_search_space = True
            if (
                isinstance(self._search_optimization_lower, Searcher)
                and hasattr(self._search_optimization_lower, "_space")
                and self._search_optimization_lower._space is not None
            ):
                if self._search_optimization_lower._metric != self._metric_name:
                    raise ValueError(
                        "If a Searcher instance has been initialized with a "
                        "space, its metric "
                        f"('{self._search_optimization_lower._metric}') "
                        "must match the metric set in TuneSearchCV"
                        f" ('{self._metric_name}')"
                    )
                if self._search_optimization_lower._mode != self.mode:
                    raise ValueError(
                        "If a Searcher instance has been initialized with a "
                        "space, its mode "
                        f"('{self._search_optimization_lower._mode}') "
                        "must match the mode set in TuneSearchCV"
                        f" ('{self.mode}')"
                    )
            elif self._is_param_distributions_all_tune_domains():
                run_args["config"].update(self.param_distributions)
                override_search_space = False

            search_kwargs = self.search_kwargs or {}
            search_kwargs = search_kwargs.copy()
            if override_search_space:
                search_kwargs["metric"] = run_args.pop("metric")
                search_kwargs["mode"] = run_args.pop("mode")
                if run_args["scheduler"]:
                    if hasattr(run_args["scheduler"], "_metric") and hasattr(
                        run_args["scheduler"], "_mode"
                    ):
                        run_args["scheduler"]._metric = search_kwargs["metric"]
                        run_args["scheduler"]._mode = search_kwargs["mode"]
                    else:
                        warnings.warn(
                            "Could not set `_metric` and `_mode` attributes "
                            f"on Scheduler {run_args['scheduler']}. "
                            "This may cause an exception later! "
                            "Ensure your Scheduler initializes with those "
                            "attributes.",
                            UserWarning,
                        )

            if self._search_optimization_lower == "bayesian":
                if override_search_space:
                    search_space = self.param_distributions
                search_algo = SkOptSearch(
                    space=search_space,
                    points_to_evaluate=best_configurations,
                    **search_kwargs,
                )
                run_args["search_alg"] = search_algo

            elif self._search_optimization_lower == "bohb":
                if override_search_space:
                    search_space = self._get_bohb_config_space()
                search_algo = TuneBOHB(
                    space=search_space,
                    seed=self.seed,
                    points_to_evaluate=best_configurations,
                    **search_kwargs,
                )
                run_args["search_alg"] = search_algo

            elif self._search_optimization_lower == "optuna":
                if "sampler" not in search_kwargs:
                    module = importlib.import_module("optuna.samplers")
                    search_kwargs["sampler"] = getattr(module, "TPESampler")(
                        seed=self.seed
                    )
                elif self.seed:
                    warnings.warn("'seed' is not implemented for Optuna.")
                if override_search_space:
                    search_space = self._get_optuna_params()
                search_algo = OptunaSearch(
                    space=search_space,
                    points_to_evaluate=best_configurations,
                    **search_kwargs,
                )
                run_args["search_alg"] = search_algo

            elif self._search_optimization_lower == "hyperopt":
                if override_search_space:
                    search_space = self._get_hyperopt_params()
                search_algo = HyperOptSearch(
                    space=search_space,
                    random_state_seed=self.seed,
                    points_to_evaluate=best_configurations,
                    **search_kwargs,
                )
                run_args["search_alg"] = search_algo

            elif isinstance(self._search_optimization_lower, Searcher):
                search_algo = self._search_optimization_lower
                run_args["search_alg"] = search_algo

            else:
                # This should not happen as we validate the input before
                # this method. Still, just to be sure, raise an error here.
                raise ValueError(
                    "Invalid search optimizer: " f"{self._search_optimization_lower}"
                )

        if (
            isinstance(self.n_jobs, int)
            and self.n_jobs > 0
            and not self._searcher_name == "random"
        ):
            search_algo = ConcurrencyLimiter(search_algo, max_concurrent=self.n_jobs)
            run_args["search_alg"] = search_algo

        run_args = self._override_run_args_with_tune_params(run_args, tune_params)

        # Currently, there is a bug where a tqdm logger instance is left unhandled during the Ray Tune. Also, it is an overkill to log Pipeline Training for Distributed Cross Validation. Hence, turning off the verbosity for `SAILPipeline`. This does not affect the Ray Tune logs which can be set via verbose field of tune.run.
        for estimator in estimator_list:
            estimator.verbosity_level = 0

        trainable = tune.with_parameters(
            trainable, X=X, y=y, estimator_list=estimator_list, fit_params=fit_params
        )

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="fail_fast='raise' " "detected.")
            analysis = tune.run(trainable, **run_args)

        if self.keep_best_configurations > 0:
            self._best_configurations = extract_top_best_configurations(
                self.keep_best_configurations,
                self._metric_name,
                analysis.results,
                self.param_distributions,
            )

        return analysis



from sklearn.preprocessing._function_transformer import FunctionTransformer

from sklearn.preprocessing._data import Binarizer
from sklearn.preprocessing._data import KernelCenterer
from sklearn.preprocessing._data import MinMaxScaler
from sklearn.preprocessing._data import MaxAbsScaler
from sklearn.preprocessing._data import Normalizer
from sklearn.preprocessing._data import RobustScaler
from sklearn.preprocessing._data import StandardScaler
from sklearn.preprocessing._data import QuantileTransformer
from sklearn.preprocessing._data import add_dummy_feature
from sklearn.preprocessing._data import binarize
from sklearn.preprocessing._data import normalize
from sklearn.preprocessing._data import scale
from sklearn.preprocessing._data import robust_scale
from sklearn.preprocessing._data import maxabs_scale
from sklearn.preprocessing._data import minmax_scale
from sklearn.preprocessing._data import quantile_transform
from sklearn.preprocessing._data import power_transform
from sklearn.preprocessing._data import PowerTransformer
from sklearn.preprocessing import PolynomialFeatures

from sklearn.preprocessing._encoders import OneHotEncoder
from sklearn.preprocessing._encoders import OrdinalEncoder

from sklearn.preprocessing._label import label_binarize
from sklearn.preprocessing._label import LabelBinarizer
from sklearn.preprocessing._label import LabelEncoder
from sklearn.preprocessing._label import MultiLabelBinarizer

from sklearn.preprocessing._discretization import KBinsDiscretizer


__all__ = [
    "Binarizer",
    "FunctionTransformer",
    "KBinsDiscretizer",
    "KernelCenterer",
    "LabelBinarizer",
    "LabelEncoder",
    "MultiLabelBinarizer",
    "MinMaxScaler",
    "MaxAbsScaler",
    "QuantileTransformer",
    "Normalizer",
    "OneHotEncoder",
    "OrdinalEncoder",
    "PowerTransformer",
    "RobustScaler",
    "StandardScaler",
    "add_dummy_feature",
    "PolynomialFeatures",
    "binarize",
    "normalize",
    "scale",
    "robust_scale",
    "maxabs_scale",
    "minmax_scale",
    "label_binarize",
    "quantile_transform",
    "power_transform",
]



"""
Classifier for ensemble models
Based on: https://github.com/yzhao062/combo/blob/master/combo/models/base.py extended for
parallelizable, incremental and batch learning algorithms.
"""

import typing
from river import base
from river import optim
from sail.models.ensemble.base import BaseAggregator
from river.compat import convert_river_to_sklearn
import ray
import numpy as np
from sail.utils.ray_utils import _model_fit

__all__ = ["DistAggregateClassifier"]


class DistAggregateClassifier(BaseAggregator):
    def __init__(
        self,
        estimators: typing.List[base.Classifier],
        fitted_estimators: typing.List[base.Classifier] = [],
        learning_rate=0.5,
        aggregator="maximization",
    ):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param aggregator: Type of aggregator. Options are "maximization" and "majority_vote"
        """
        # if len(estimators) < 2:
        #     raise NotEnoughModels(n_expected=2, n_obtained=len(regressors))

        self.learning_rate = learning_rate
        self.weights = [1.0] * len(estimators)
        self.aggregator = aggregator
        estimators = [
            convert_river_to_sklearn(est) if "river" in est.__module__ else est
            for est in estimators
        ]
        super().__init__(
            base_estimators=estimators, fitted_estimators=fitted_estimators
        )

    def _partial_fit(self, X, y=None, classes=None, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        obj_ids = []
        for i, estimator in enumerate(self.base_estimators):
            estimator.n_features_in_ = X.shape[1]
            X_id = ray.put(X)
            obj_ids.append(_model_fit.remote(estimator, X_id, y, classes))
        estimators = ray.get(obj_ids)
        self.base_estimators = estimators
        return self

    def partial_fit(self, X, y=None, classes=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        self._partial_fit(X, y, classes)
        return self

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Class labels/predictions for input samples.
        """
        X_ensemble = []
        estimators = self.base_estimators
        estimators.extend(self.fitted_estimators)
        for i, estimator in enumerate(estimators):
            estimator.n_features_in_ = X.shape[1]
            X_ensemble.append(estimator.predict(X))
        X_ensemble = np.array(X_ensemble)

        # maximization
        if self.aggregator == "maximization":
            y_pred = np.max(X_ensemble, axis=1).ravel()
        # majority vote
        if self.aggregator == "majority_vote":
            n_samples, n_estimators = X_ensemble.shape[0], X_ensemble.shape[1]
            vote_results = np.zeros(
                [
                    n_samples,
                ]
            )
            for i in range(n_samples):
                values, counts = np.unique(X_ensemble[i], return_counts=True)
                ind = np.argmax(counts)
                vote_results[i] = values[ind]
            y_pred = vote_results.ravel()
            y_pred = np.array([int(i) for i in y_pred])
        # # average
        # if self.aggregator == "average":
        #     y_pred = np.nanmean(X_ensemble, axis=1).ravel()

        return y_pred



import typing
from river import neighbors
from sail.models.river.base import SailRiverClassifier, SailRiverRegressor

__all__ = ["NearestNeighbors", "KNNClassifier", "KNNRegressor"]


class NearestNeighbors(SailRiverClassifier):
    def __init__(
        self,
        window_size: int,
        min_distance_keep: float,
        distance_func: typing.Union[
            neighbors.base.DistanceFunc, neighbors.base.FunctionWrapper
        ],
    ):
        super(NearestNeighbors, self).__init__(
            river_estimator=neighbors.NearestNeighbors(
                window_size,
                min_distance_keep,
                distance_func,
            )
        )


class KNNClassifier(SailRiverClassifier):
    def __init__(
        self,
        n_neighbors: int = 5,
        window_size: int = 1000,
        leaf_size: int = 30,
        p: float = 2,
        weighted: bool = True,
        **kwargs,
    ):
        super(KNNClassifier, self).__init__(
            river_estimator=neighbors.KNNClassifier(
                n_neighbors, window_size, leaf_size, p, weighted, **kwargs
            )
        )


class KNNRegressor(SailRiverRegressor):
    def __init__(
        self,
        n_neighbors: int = 5,
        window_size: int = 1000,
        leaf_size: int = 30,
        p: float = 2,
        aggregation_method: str = "mean",
        **kwargs,
    ):
        super(KNNRegressor, self).__init__(
            river_estimator=neighbors.KNNRegressor(
                n_neighbors, window_size, leaf_size, p, aggregation_method, **kwargs
            )
        )



from sklearn.base import OneToOneFeatureMixin

from .base import SAILTransformer


class ColumnNamePrefixTransformer(OneToOneFeatureMixin, SAILTransformer):
    def __init__(self, prefix="") -> None:
        super(ColumnNamePrefixTransformer, self).__init__()
        self.prefix = prefix
        self.validation_params["cast_to_ndarray"] = False

    def _partial_fit(self, X, y=None):
        return self

    def _transform(self, X, y=None, copy=None):
        if copy:
            new_df = X.copy()
        else:
            new_df = X

        new_columns = []
        for feature in X.columns:
            new_columns.append(self.prefix + "_" + feature)

        new_df.columns = new_columns
        self.feature_names_in_ = new_columns

        return new_df.to_numpy()



from sklearn import decomposition
import numpy as np


class PCA(decomposition.PCA):
    def __init__(
        self,
        prefix: str,
        n_components=None,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        n_oversamples=10,
        power_iteration_normalizer="auto",
        random_state=None,
    ) -> None:
        self.prefix = prefix
        super().__init__(
            n_components,
            copy=copy,
            whiten=whiten,
            svd_solver=svd_solver,
            tol=tol,
            iterated_power=iterated_power,
            n_oversamples=n_oversamples,
            power_iteration_normalizer=power_iteration_normalizer,
            random_state=random_state,
        )

    def get_feature_names_out(self, input_features=None):
        """Get output feature names for transformation.

        The feature names out will prefixed by the lowercased class name. For
        example, if the transformer outputs 3 features, then the feature names
        out are: `["class_name0", "class_name1", "class_name2"]`.

        Parameters
        ----------
        input_features : array-like of str or None, default=None
            Only used to validate feature names with the names seen in :meth:`fit`.

        Returns
        -------
        feature_names_out : ndarray of str objects
            Transformed feature names.
        """
        return np.asarray(
            [f"{self.prefix}_pca{i}" for i in range(self._n_features_out)], dtype=object
        )



# PyTorch implementation for LSTM FCN for Time Series Classification without LSTM
# Original code in TensorFlow https://github.com/titu1994/LSTM-FCN
# Paper https://arxiv.org/abs/1709.05206
#
# By David Campos and Teodor Vernica

import torch, skorch
from torch import nn
from skorch.classifier import NeuralNetClassifier
from sail.models.torch.layers import ConvBlock # The convolutional block using Conv1D with same padding.


class _FCN(nn.Module):

    def __init__(self, in_channels: int, input_size: int, lstm_layers: int = 8, classes: int = 1) -> None:
        """FCN model.
        Args:
            in_channels (int): Number of input channels
            input_size (int): The input size
            lstm_layers (int): Number of hidden LSTM units
            classes (int): Number of classes
        """
        super().__init__()
        
        self.conv_layers = nn.Sequential(*[
            ConvBlock(in_channels, 128, 8, 1),
            ConvBlock(128, 256, 5, 1),
            ConvBlock(256, 128, 3, 1),
        ])

        self.fc = nn.Linear(128, classes)
        self.softmax = nn.Softmax(dim=1)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.unsqueeze(x, 2)
        
        x_cnn = self.conv_layers(x.transpose(2,1))
        x_cnn = torch.mean(x_cnn,dim=-1)

        x_out = self.softmax(self.fc(x_cnn))

        return x_out

    
class FCN_Classifier(NeuralNetClassifier):
    def __init__(self, in_channels, input_size, lstm_layers, classes):
        """FCN model.
        Args:
            in_channels (int): Number of input channels
            input_size (int): The input size
            lstm_layers (int): Number of hidden LSTM units
            classes (int): Number of classes
        """
        super(FCN_Classifier, self).__init__(
            module=_FCN,
            module__in_channels=in_channels,
            module__input_size=input_size,
            module__lstm_layers=lstm_layers,
            module__classes=classes,
            max_epochs=1)


import torch
from torch import nn
import torch.nn.functional as F

"""Represents the "Same" padding functionality from Tensorflow.
    See: https://github.com/pytorch/pytorch/issues/3867"""
class Conv1dSamePadding(nn.Conv1d):
    def forward(self, input):
        return conv1d_same_padding(input, self.weight, self.bias, self.stride,
                                   self.dilation, self.groups)


def conv1d_same_padding(input, weight, bias, stride, dilation, groups):
    kernel, dilation, stride = weight.size(2), dilation[0], stride[0]
    l_out = l_in = input.size(2)
    padding = (((l_out - 1) * stride) - l_in + (dilation * (kernel - 1)) + 1)
    if padding % 2 != 0:
        input = F.pad(input, [0, 1])

    return F.conv1d(input=input, weight=weight, bias=bias, stride=stride,
                    padding=padding // 2,
                    dilation=dilation, groups=groups)


"""Convolutional 1D block according to the paper specification"""
class ConvBlock(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, kernel_size: int,
                 stride: int) -> None:
        super().__init__()

        self.conv_layers = nn.Sequential(
            Conv1dSamePadding(in_channels=in_channels,out_channels=out_channels,
                              kernel_size=kernel_size,stride=stride),
            nn.BatchNorm1d(num_features=out_channels),
            nn.ReLU(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.conv_layers(x)


from typing import Dict, List, Optional
from tune_sklearn.list_searcher import ListSearcher, RandomListSearcher
from ray.tune.utils.util import _atomic_save, _load_newest_checkpoint


class SearcherSerializationMixin:
    def get_state(self):
        state = self.__dict__.copy()
        return state

    def set_state(self, state):
        self.__dict__.update(state)

    def save_to_dir(self, dirpath, session_str):
        state_dict = self.get_state()
        _atomic_save(
            state=state_dict,
            checkpoint_dir=dirpath,
            file_name=self.CKPT_FILE_TMPL.format(session_str),
            tmp_file_name=".tmp_generator",
        )

    def restore_from_dir(self, dirpath: str):
        """Restores self + searcher + search wrappers from dirpath."""
        state_dict = _load_newest_checkpoint(dirpath, self.CKPT_FILE_TMPL.format("*"))
        if not state_dict:
            raise RuntimeError("Unable to find checkpoint in {}.".format(dirpath))
        self.set_state(state_dict)


class SailListSearcher(SearcherSerializationMixin, ListSearcher):
    """Custom search algorithm to support passing in a list of
    dictionaries to SailTuneGridSearchCV

    """

    CKPT_FILE_TMPL = "sail-list-searcher-state-{}.json"

    def __init__(self, param_grid, points_to_evaluate: Optional[List[Dict]] = None):
        super(SailListSearcher, self).__init__(param_grid)
        self._points_to_evaluate = points_to_evaluate
        self._setup()

    def _setup(self) -> None:
        if self._points_to_evaluate is None:
            self._best_configurations = []
        else:
            self._best_configurations = self._points_to_evaluate

    def suggest(self, trial_id):
        if self._best_configurations:
            return self._best_configurations.pop(0)
        else:
            return super().suggest(trial_id)


class SailRandomListSearcher(SearcherSerializationMixin, RandomListSearcher):
    """Custom search algorithm to support passing in a list of
    dictionaries to SailTuneSearchCV

    """

    CKPT_FILE_TMPL = "random-list-searcher-state-{}.json"

    def __init__(self, param_grid, points_to_evaluate: Optional[List[Dict]] = None):
        super(SailRandomListSearcher, self).__init__(param_grid)
        self._points_to_evaluate = points_to_evaluate
        self._setup()

    def _setup(self) -> None:
        if self._points_to_evaluate is None:
            self._best_configurations = []
        else:
            self._best_configurations = self._points_to_evaluate

    def suggest(self, trial_id):
        if self._best_configurations:
            return self._best_configurations.pop(0)
        else:
            return super().suggest(trial_id)



import torch
import torch.nn as nn
import torch.nn.functional as F
from skorch import NeuralNetClassifier
from torch.nn.parameter import Parameter


class _ONNHBPModel(nn.Module):
    """Online Neural Network with Hedge BackPropagation.
    Online Deep Learning: Learning Deep Neural Networks on the Fly https://arxiv.org/abs/1711.03705

    Paper https://arxiv.org/abs/1711.03705
    Original code in Theano https://github.com/LIBOL/ODL
    Based on Pytorch implementation https://github.com/alison-carrera/onn

    Implemented by: Carlos Muniz Cuza and Jonas Brusokas

        Args:
            input_units: dimensionality of input
            output_units: number of classification classes
            hidden_units: number of hidden units in a single layer
            n_hidden_layers: number of layers in the network
            dropout: dropout coefficient
            beta: beta coefficient (b)
            learning_rate: learning rate for training
            smoothing: smoothing coefficient (s)
    """

    def __init__(self,
                 input_units: int,
                 output_units: int,
                 hidden_units: int,
                 n_hidden_layers: int = 1,
                 dropout: float = 0.2,
                 beta: float = 0.99,
                 learning_rate: float = 0.01,
                 smoothing: float = 0.2,
                 activation: str = 'ReLU',
                 device: str = 'cpu'
                 ):
        super(_ONNHBPModel, self).__init__()

        self.input_units = input_units
        self.output_units = output_units
        self.hidden_units = hidden_units
        self.n_hidden_layers = n_hidden_layers
        # self.batch_size = batch_size

        self.device = torch.device(device)

        self.beta = Parameter(torch.tensor(beta), requires_grad=False).to(self.device)
        self.learning_rate = Parameter(torch.tensor(learning_rate), requires_grad=False).to(self.device)
        self.smoothing = Parameter(torch.tensor(smoothing), requires_grad=False).to(self.device)

        self.loss_array = []

        self.hidden_layers = nn.ModuleList(
            [nn.Linear(input_units, hidden_units)] +
            [nn.Linear(hidden_units, hidden_units) for _ in range(self.n_hidden_layers - 1)])  #

        self.output_layers = nn.ModuleList([nn.Linear(hidden_units, output_units) for _ in range(
            self.n_hidden_layers)])  #

        self.alpha = Parameter(torch.Tensor(self.n_hidden_layers).fill_(1 / (self.n_hidden_layers + 1)),
                               requires_grad=False)  #

        self.do = nn.Dropout(p=dropout)

        try:
            self.actfn = getattr(nn, activation)()
        except AttributeError:
            print('Invalid activation function: choose ReLu by default')
            self.actfn = nn.ReLU

        self.dtype = torch.float  # ?

    def zero_grad_(self):  #
        for i in range(self.n_hidden_layers):
            self.output_layers[i].weight.grad.data.fill_(0)
            self.output_layers[i].bias.grad.data.fill_(0)
            self.hidden_layers[i].weight.grad.data.fill_(0)
            self.hidden_layers[i].bias.grad.data.fill_(0)

    def __update_alpha(self, losses_per_layer, l: int):  #
        self.alpha[l] *= torch.pow(self.beta, losses_per_layer[l])
        self.alpha[l] = torch.max(self.alpha[l], self.smoothing / self.n_hidden_layers)

    def __update_hidden_layer_weight(self, w, b, l):  #
        self.hidden_layers[l].weight.data -= self.learning_rate * w
        self.hidden_layers[l].bias.data -= self.learning_rate * b

    def calculate_CE_loss(self, X, Y):
        """
        Helper method to calculate the Cross Entropy (CE) loss given ground-truth X and Y
        Args:
            X: ground-truth input
            Y: ground-truth labels/classes

        Returns:
            mean_loss: mean CE loss across all layers
            losses_per_layer: list of CE losses per layer
        """
        predictions_per_layer = self.forward_(X)

        losses_per_layer = []
        for out in predictions_per_layer:
            criterion = nn.CrossEntropyLoss().to(self.device)
            # loss = criterion(out.view(batch_size, n_classes), Y.view(batch_size).long())
            loss = criterion(out, Y.long())
            losses_per_layer.append(loss)

        mean_loss = torch.stack(losses_per_layer).mean().detach()
        return mean_loss, losses_per_layer

    def update_weights(self, X, Y):
        # batch_size = Y.shape
        # n_classes = self.output_units
        if not isinstance(Y, torch.Tensor):
            Y = torch.from_numpy(Y).to(self.device)
        total_predictions_before_update = self.predict_(X)

        mean_loss, losses_per_layer = self.calculate_CE_loss(X, Y)

        w = [0] * self.n_hidden_layers
        b = [0] * self.n_hidden_layers

        with torch.no_grad():
            for i in range(self.n_hidden_layers):
                losses_per_layer[i].backward(retain_graph=True)
                self.output_layers[i].weight.data -= self.learning_rate * self.alpha[i] * self.output_layers[
                    i].weight.grad.data
                self.output_layers[i].bias.data -= self.learning_rate * self.alpha[i] * self.output_layers[
                    i].bias.grad.data

                for j in range(i + 1):
                    w[j] += self.alpha[i] * self.hidden_layers[j].weight.grad.data
                    b[j] += self.alpha[i] * self.hidden_layers[j].bias.grad.data

                self.zero_grad_()

            [self.__update_hidden_layer_weight(w[i], b[i], i) for i in range(self.n_hidden_layers)]
            [self.__update_alpha(losses_per_layer, i) for i in range(self.n_hidden_layers)]

        z_t = torch.sum(self.alpha)
        self.alpha = Parameter(self.alpha / z_t, requires_grad=False).to(self.device)

        # Return the averaged loss across layers (maybe 'sum' could be more appropriate?)
        return total_predictions_before_update, mean_loss

    def forward(self, X: torch.Tensor):
        scores = self.predict_(X_data=X)
        return scores

    # NOTE: we do not have the 'show_loss' here
    def partial_fit_(self, X_data, Y_data):
        # self.validate_input_X(X_data)
        # self.validate_input_Y(Y_data)
        return self.update_weights(X_data, Y_data)

    def partial_fit(self, X_data, Y_data):
        return self.partial_fit_(X_data, Y_data)

    def forward_(self, X: torch.Tensor):
        """
        Args:
            X: input data
        Returns: [layers, batch, classes] tensor with predictions from each layer
        """
        if not isinstance(X, torch.Tensor):
            X = torch.from_numpy(X)
        X = X.float().to(self.device)
        x = F.relu(self.hidden_layers[0](X))

        hidden_connections = [x]

        for i in range(1, self.n_hidden_layers):
            hidden_connections.append(F.relu(self.hidden_layers[i](hidden_connections[i - 1])))

        output_class = []
        for i in range(self.n_hidden_layers):
            output_class.append(self.output_layers[i](hidden_connections[i]))

        pred_per_layer = torch.stack(output_class)
        return pred_per_layer

    def predict_(self, X_data):
        """
        Args:
            X_data: input data
        Returns: [batch, class_count] tensor with scores for each class
        """
        scores = torch.sum(
            torch.mul(
                self.alpha.view(self.n_hidden_layers, 1).repeat(1, len(X_data))
                    .view(self.n_hidden_layers, len(X_data), 1)
                , self.forward_(X_data))
            , 0)
        return scores.softmax(dim=1)

    def predict(self, X_data):
        """
        Args:
            X_data: input data
        Returns: [batch] tensor with argmax'ed predictions (MAP) for each input item
        """
        scores = self.predict_(X_data)
        return torch.argmax(scores, dim=1)


class ONNHBPClassifier(NeuralNetClassifier):
    def __init__(self,
                 # in_channels, input_size, lstm_layers, classes,
                 input_units: int,
                 output_units: int,
                 hidden_units: int,
                 n_hidden_layers: int = 1,
                 dropout: float = 0.2,
                 beta: float = 0.99,
                 learning_rate: float = 0.01,
                 smoothing: float = 0.2,
                 activation: str = 'ReLU',
                 **kwargs
                 ):
        """
        Online Neural Network trained with Hedge BackPropagation Classifier.

        Paper https://arxiv.org/abs/1711.03705
        Original code in Theano https://github.com/LIBOL/ODL
        Based on Pytorch implementation https://github.com/alison-carrera/onn

        Implemented by: Carlos Muniz Cuza and Jonas Brusokas

        Args:
            input_units: dimensionality of input
            output_units: number of classification classes
            hidden_units: number of hidden units in a single layer
            n_hidden_layers: number of layers in the network
            dropout: dropout coefficient
            beta: beta coefficient (b)
            learning_rate: learning rate for training
            smoothing: smoothing coefficient (s)
            kwargs: NeuralNetClassifier parameters
        """
        super(ONNHBPClassifier, self).__init__(
            module=_ONNHBPModel,
            module__input_units=input_units,
            module__output_units=output_units,
            module__hidden_units=hidden_units,
            module__n_hidden_layers=n_hidden_layers,
            module__dropout=dropout,
            module__beta=beta,
            module__learning_rate=learning_rate,
            module__smoothing=smoothing,
            module__activation=activation,
            max_epochs=1,
            **kwargs
        )

    # OVERRIDES net.py:965
    # This enables training the model EXACTLY as specified in the paper
    # and avoiding all of the defaults imposed by 'skorch'
    def train_step(self, batch, **fit_params):
        step_accumulator = self.get_train_step_accumulator()
        y_pred, loss = self.module_.partial_fit(batch[0], batch[1])
        step = {
            "y_pred": y_pred,
            "loss": loss
        }
        step_accumulator.store_step(step)
        return step_accumulator.get_step()

    # This enables logging "valid_loss"
    def validation_step(self, batch, **fit_params):
        val_loss, _ = self.module_.calculate_CE_loss(batch[0], batch[1])
        y_pred = self.module_.predict_(batch[0])
        return {
            'loss': val_loss,
            'y_pred': y_pred,
        }


if __name__ == '__main__':
    from sklearn.model_selection import train_test_split
    from sklearn.datasets import make_classification
    import numpy as np

    X, Y = make_classification(n_samples=5000, n_features=10, n_informative=4, n_redundant=0, n_classes=10,
                               n_clusters_per_class=1, class_sep=3)

    X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=0.3, random_state=42, shuffle=True)

    print('Training dataset size', X_train.shape)
    print('Testing dataset size', X_test.shape)

    onn_network = ONNHBPClassifier(input_units=10,
                                   output_units=10,
                                   hidden_units=40,
                                   n_hidden_layers=5,
                                   # activation='Tanh',
                                   train_split=None,
                                   verbose=0
                                   )

    partial_fit = onn_network.partial_fit(np.asarray([X_train[0, :]]), np.asarray([y_train[0]]))

    print('Model', partial_fit)

    print("Online Accuracy at the beginning {}".format(partial_fit.score(X_test, y_test)))

    print('Training on more examples ...')
    for i in range(1, 1000):
        partial_fit = onn_network.partial_fit(np.asarray([X_train[i, :]]), np.asarray([y_train[i]]))

    print("Online Accuracy after 1000 examples: {}".format(partial_fit.score(X_test, y_test)))




"""
Regressor for ensemble models
Based on: https://github.com/yzhao062/combo/blob/master/combo/models/base.py extended for
parallelizable incremental and batch learning algorithms.
"""
import typing
from river import base
from river import optim
from .base import BaseAggregator
from river.compat import convert_river_to_sklearn
import ray
import numpy as np
from sail.utils.ray_utils import _model_fit


__all__ = ["DistAggregateRegressor"]


class DistAggregateRegressor(BaseAggregator):

    def __init__(
        self,
        estimators,
        fitted_estimators = [],
        loss = None,
        learning_rate=0.5,
        aggregator="simple"
    ):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param aggregator: Type of aggregator. Options are "simple", "windsor" and "trim"
        """
        # if len(estimators) < 2:
        #     raise NotEnoughModels(n_expected=2, n_obtained=len(estimators))

        self.loss = optim.losses.Squared() if loss is None else loss
        self.learning_rate = learning_rate
        self.weights = [1.0] * len(estimators)
        self.aggregator = aggregator
        estimators = [convert_river_to_sklearn(est) if "river" in est.__module__ else est
                      for est in estimators]
        # self.estimators = [est.remote() for est in estimators]
        super().__init__(base_estimators=estimators, fitted_estimators=fitted_estimators)

    def _partial_fit(self, X, y=None, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return:
        """
        obj_ids = []

        for i, estimator in enumerate(self.base_estimators):
            estimator.n_features_in_ = X.shape[1]
            X_id = ray.put(X)
            obj_ids.append(_model_fit.remote(estimator, X_id, y))
        self.base_estimators = ray.get(obj_ids)
        return self

    def partial_fit(self, X, y=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :return:
        """
        self._partial_fit(X, y)
        return self

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Predictions for input samples.
        """
        X_ensemble = []
        self.estimators = self.base_estimators
        self.estimators.extend(self.fitted_estimators)
        X_ensemble.append([model.predict(X) for model in self.estimators])
        X_ensemble = np.array(X_ensemble)
        # simple
        if self.aggregator == "simple":
            y_pred = np.nanmean(X_ensemble, axis=1)[0]
        # Windsorized
        if self.aggregator == "windsor":
            y_pred = ((np.sum(X_ensemble, axis=1) - np.amax(X_ensemble, axis=1) - np.amin(X_ensemble, axis=1)
                       + np.sort(X_ensemble)[:, 1] + np.sort(X_ensemble)[:, -2]) / (
                    X_ensemble.shape[1]))[0]
        # Trimmed
        if self.aggregator == "trim":
            y_pred = ((np.sum(X_ensemble, axis=1) - np.amax(X_ensemble, axis=1) -
                      np.amin(X_ensemble, axis=1)) / (X_ensemble.shape[1] - 2))[0]
        return y_pred


from sklearn.utils import check_array, check_X_y


def log_epoch(func):
    def wrapper(*args, **kwargs):
        estimator = args[0]
        estimator.verbosity.log_epoch(args[1])
        func(*args, **kwargs)

    return wrapper


def validate_X_y(func):
    def wrapper(*args, **kwargs):
        X = args[1]
        y = args[2]

        fit_params = {}
        if len(args) == 4:
            fit_params = args[3]

        if len(X.shape) == 1:
            X = X.reshape((1, X.shape[0]))

        X_new = X.copy()
        datetime_cols = list(X_new.select_dtypes(include="datetime64[ns]"))
        X_new = X_new.drop(datetime_cols, axis=1)

        if y is None:
            _ = check_array(
                X_new,
                dtype=None,
                input_name="X",
            )
        else:
            _, y = check_X_y(X_new, y, dtype=None)

        func(args[0], X=X, y=y, **fit_params)

    return wrapper



__version__ = "0.10.0"
__package_name__ = "SAIL"
_logging_level = "INFO"


def set_logging_level(logging_level):
    global _logging_level
    _logging_level = logging_level


def get_logging_level():
    return _logging_level



import river.feature_extraction.agg as agg
import river.feature_extraction.kernel_approx as kernel_approx
import river.feature_extraction.poly as poly
import river.feature_extraction.vectorize as vectorize
from sail.transformers.river.base import BaseRiverTransformer
import typing
from river import stats, utils
from sklearn.base import ClassNamePrefixFeaturesOutMixin
import numpy as np

__all__ = [
    "Agg",
    "BagOfWords",
    "PolynomialExtender",
    "RBFSampler",
    "TargetAgg",
    "TFIDF",
]


class Agg(BaseRiverTransformer):
    def __init__(
        self,
        on: str,
        by: str | list[str] | None,
        how: stats.base.Univariate | utils.Rolling | utils.TimeRolling,
    ):
        super(Agg, self).__init__(
            river_estimator=agg.Agg(
                on,
                by,
                how,
            )
        )


class TargetAgg(ClassNamePrefixFeaturesOutMixin, BaseRiverTransformer):
    def __init__(
        self,
        by: str | list[str] | None,
        how: stats.base.Univariate | utils.Rolling | utils.TimeRolling,
        target_name="y",
    ):
        super(TargetAgg, self).__init__(
            river_estimator=agg.TargetAgg(
                by,
                how,
                target_name,
            )
        )
        self.validation_params["cast_to_ndarray"] = False

    def get_feature_names_out(self, input_features=None):
        """Get output feature names for transformation.

        The feature names out will prefixed by the lowercased class name. For
        example, if the transformer outputs 3 features, then the feature names
        out are: `["class_name0", "class_name1", "class_name2"]`.

        Parameters
        ----------
        input_features : array-like of str or None, default=None
            Only used to validate feature names with the names seen in :meth:`fit`.

        Returns
        -------
        feature_names_out : ndarray of str objects
            Transformed feature names.
        """
        self._feature_name = f"{self.on}_{self.how.name}"
        if self.by:
            self._feature_name += f"_by_{'_and_'.join(self.by)}"
        return np.asarray([self._feature_name], dtype=object)


class RBFSampler(BaseRiverTransformer):
    def __init__(self, gamma=1.0, n_components=100, seed: int | None = None):
        super(RBFSampler, self).__init__(
            river_estimator=kernel_approx.RBFSampler(gamma, n_components, seed)
        )


class PolynomialExtender(BaseRiverTransformer):
    def __init__(
        self, degree=2, interaction_only=False, include_bias=False, bias_name="bias"
    ):
        super(PolynomialExtender, self).__init__(
            river_estimator=poly.PolynomialExtender(
                degree, interaction_only, include_bias, bias_name
            )
        )


class TFIDF(BaseRiverTransformer):
    def __init__(
        self,
        normalize=True,
        on: str | None = None,
        strip_accents=True,
        lowercase=True,
        preprocessor: typing.Callable | None = None,
        tokenizer: typing.Callable | None = None,
        ngram_range=(1, 1),
    ):
        super(TFIDF, self).__init__(
            river_estimator=vectorize.TFIDF(
                normalize,
                on,
                strip_accents,
                lowercase,
                preprocessor,
                tokenizer,
                ngram_range,
            )
        )


class BagOfWords(BaseRiverTransformer):
    def __init__(
        self,
        on: str | None = None,
        strip_accents=True,
        lowercase=True,
        preprocessor: typing.Callable | None = None,
        stop_words: set[str] | None = None,
        tokenizer_pattern=r"(?u)\b\w[\w\-]+\b",
        tokenizer: typing.Callable | None = None,
        ngram_range=(1, 1),
    ):
        super(BagOfWords, self).__init__(
            river_estimator=vectorize.BagOfWords(
                on,
                strip_accents,
                lowercase,
                preprocessor,
                stop_words,
                tokenizer_pattern,
                tokenizer,
                ngram_range,
            )
        )



import ray
import numpy as np
import numpy.typing as npt

"""
Ray functions for distributed training and scoring
"""

@ray.remote
def _model_fit(model: object, X: npt.NDArray, y: npt.ArrayLike, classes: list = None):
    """
    :param model: Any machine learning model with partial_fit function defined.
    :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
    :param y: numpy.ndarray of shape (n_samples)
           Labels for the target variable.
    :param classes: numpy.ndarray, optional.
           Unique classes in the data y.
    :return: Trained machine learning model.
    """
    if classes is None:
        trained_model = model.partial_fit(X, y)
    else:
        trained_model = model.partial_fit(X, y, classes)
    return trained_model


@ray.remote
def _model_fit_classifier(model: object, X: npt.NDArray, y: npt.ArrayLike, classes: list = None):
    if classes is None:
        classes = np.unique(y)
    trained_model = model.partial_fit(X, y, classes)
    return trained_model


@ray.remote
def _model_predict(model: object, X: npt.NDArray):
    """
    :param model: Any machine learning model with partial_fit function defined.
    :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
    :return: numpy.ndarray of shape (n_samples)
        Predictions for the samples in X.
    """
    return model.predict(X)


@ray.remote
def _model_metrics(model: object, X: npt.NDArray, y_true: npt.ArrayLike, metric: object):
    """
    :param model: Any machine learning model with partial_fit function defined.
    :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
    :param y_true: numpy.ndarray of shape (n_samples)
           True labels or ground truth for the samples in X
    :param metric: Object that defines the quality of predictions
           (ex. metrics.accuracy_score in scikit-learn)
    :return: Metric as defined by the metric object.
    """
    y_pred = model.predict(X)
    return metric(y_true, y_pred)



from sklearn import compose
import numpy as np
from scipy import sparse
from sklearn.utils import _print_elapsed_time


def _partial_fit_transform_one(
    transformer, X, y, weight, message_clsname="", message=None, **fit_params
):
    """
    Fits ``transformer`` to ``X`` and ``y``. The transformed result is returned
    with the fitted transformer. If ``weight`` is not ``None``, the result will
    be multiplied by ``weight``.
    """
    with _print_elapsed_time(message_clsname, message):
        if hasattr(transformer, "partial_fit_transform"):
            res = transformer.partial_fit_transform(X, y, **fit_params)
        elif hasattr(transformer, "partial_fit"):
            res = transformer.partial_fit(X, y, **fit_params).transform(X)
        else:
            res = transformer.fit(X, y, **fit_params).transform(X)

    if weight is None:
        return res, transformer
    return res * weight, transformer


class ColumnTransformer(compose.ColumnTransformer):
    def partial_fit(self, X, y=None):
        """Fit all transformers using X.

        Parameters
        ----------
        X : {array-like, dataframe} of shape (n_samples, n_features)
            Input data, of which specified subsets are used to fit the
            transformers.

        y : array-like of shape (n_samples,...), default=None
            Targets for supervised learning.

        Returns
        -------
        self : ColumnTransformer
            This estimator.
        """
        # we use fit_transform to make sure to set sparse_output_ (for which we
        # need the transformed data) to have consistent output type in predict
        self.partial_fit_transform(X, y=y)
        return self

    def partial_fit_transform(self, X, y=None):
        """Fit all transformers, transform the data and concatenate results.

        Parameters
        ----------
        X : {array-like, dataframe} of shape (n_samples, n_features)
            Input data, of which specified subsets are used to fit the
            transformers.

        y : array-like of shape (n_samples,), default=None
            Targets for supervised learning.

        Returns
        -------
        X_t : {array-like, sparse matrix} of \
                shape (n_samples, sum_n_components)
            Horizontally stacked results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers. If
            any result is a sparse matrix, everything will be converted to
            sparse matrices.
        """
        first_call = not hasattr(self, "n_features_in_")
        self._check_feature_names(X, reset=first_call)

        X = compose._column_transformer._check_X(X)
        # set n_features_in_ attribute
        self._check_n_features(X, reset=first_call)
        self._validate_transformers()
        self._validate_column_callables(X)
        self._validate_remainder(X)

        result = self._fit_transform(
            X, y, _partial_fit_transform_one, fitted=not first_call
        )

        if not result:
            self._update_fitted_transformers([])
            # All transformers are None
            return np.zeros((X.shape[0], 0))

        Xs, transformers = zip(*result)

        # determine if concatenated output will be sparse or not
        if any(sparse.issparse(X) for X in Xs):
            nnz = sum(X.nnz if sparse.issparse(X) else X.size for X in Xs)
            total = sum(
                X.shape[0] * X.shape[1] if sparse.issparse(X) else X.size for X in Xs
            )
            density = nnz / total
            self.sparse_output_ = density < self.sparse_threshold
        else:
            self.sparse_output_ = False

        self._update_fitted_transformers(transformers)
        self._validate_output(Xs)
        self._record_output_indices(Xs)

        return self._hstack(list(Xs))



from sklearn.metrics import accuracy_score

from .base import ModelSelectorBase


class HoldoutBestModelSelector(ModelSelectorBase):
    def __init__(
        self, estimators, fitted_estimators=[], metrics=accuracy_score
    ):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param fitted_estimators: Batch learning estimators.
                Used only for scoring.
        :param metrics: Object that defines the quality of predictions
           (ex. metrics.accuracy_score in scikit-learn)
        """
        self.best_model_index = 0
        super().__init__(
            estimators=estimators,
            fitted_estimators=fitted_estimators,
            metrics=metrics,
        )

    def partial_fit(self, X, y=None, classes=None):
        """
        :param model: Any machine learning model with partial_fit function defined.
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param classes: numpy.ndarray, optional.
               Unique classes in the data y.
        :return:
        """
        self._partial_fit_estimators(X, y, classes)
        return self

    def _get_best_model_index(self, X_test, y_test):
        """
        :param X_test: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y_test: numpy.ndarray of shape (n_samples)
               Labels (ground truth) for the target variable.
        :return: int. Returns the index of the best model based on user defined metrics.
        """
        scores = []
        estimators = self.base_estimators
        estimators.extend(self.fitted_estimators)
        for estimator in estimators:
            estimator.n_features_in_ = X_test.shape[1]
            # scores.append(estimator.predict(X))
            scores.append(self.metrics(y_test, estimator.predict(X_test)))
        return scores.index(max(scores))



from .base import SAILTransformer
import numpy as np
import pandas as pd
from sklearn.base import ClassNamePrefixFeaturesOutMixin
from sklearn.utils.validation import check_is_fitted


class Polar2CartTransformer(ClassNamePrefixFeaturesOutMixin, SAILTransformer):
    def __init__(self, n, features=None, suffix_x="x", suffix_y="y") -> None:
        super(Polar2CartTransformer, self).__init__()
        self.n = n
        self.features = features
        self.suffix_x = suffix_x
        self.suffix_y = suffix_y
        self.validation_params["cast_to_ndarray"] = False

    def _partial_fit(self, X, y=None):
        return self

    def _transform(self, X, y=None, copy=None):
        new_df = pd.DataFrame()

        features = self.features if self.features else X.columns
        for feature in features:
            name_x = feature + "_" + self.suffix_x
            new_df[name_x] = np.cos(2 * np.pi * X[feature] / self.n)
            name_y = feature + "_" + self.suffix_y
            new_df[name_y] = np.sin(2 * np.pi * X[feature] / self.n)

        self.features_ = list(new_df.columns)
        return new_df.to_numpy()

    def get_feature_names_out(self, input_features=None):
        """Get output feature names for transformation.

        Parameters
        ----------
        input_features : array-like of str or None, default=None
            Input features.

            - If `input_features` is `None`, then `feature_names_in_` is
              used as feature names in. If `feature_names_in_` is not defined,
              then the following input feature names are generated:
              `["x0", "x1", ..., "x(n_features_in_ - 1)"]`.
            - If `input_features` is an array-like, then `input_features` must
              match `feature_names_in_` if `feature_names_in_` is defined.

        Returns
        -------
        feature_names_out : ndarray of str objects
            Same as input features.
        """
        check_is_fitted(self, "n_features_in_")
        return self.features_



"""
Weighted Gradient learning based LSTM (WGLSTM) neural network

Code adapted from https://github.com/weilai0980/onlineLearning
"""

import copy
from types import SimpleNamespace

import numpy as np
import scipy.stats as st
import tensorflow as tf
from numpy.lib.scimath import sqrt
from scikeras.wrappers import KerasRegressor

from sail.models.keras.base import KerasSerializationMixin
from sail.utils.stats import nmse


class _Model(tf.keras.models.Sequential):
    def __init__(
        self,
        num_of_features=1,
        hidden_layer_neurons=450,
        hidden_layer_activation="linear",
        regularization_factor=0.0001,
        timesteps=1,
        window_size=20,
    ):
        super(_Model, self).__init__(name="WGLSTM")
        self.window_size = window_size
        self.add(
            tf.keras.layers.LSTM(
                hidden_layer_neurons,
                return_sequences=True,
                stateful=True,
                batch_input_shape=(1, timesteps, num_of_features),
                kernel_regularizer=tf.keras.regularizers.l2(regularization_factor),
            )
        )
        self.add(
            tf.keras.layers.TimeDistributed(
                tf.keras.layers.Dense(units=num_of_features)
            )
        )
        self.add(tf.keras.layers.Activation(hidden_layer_activation))

    def fit(self, x, y, **kwargs):
        """
        Trains the model for a fixed number of epochs
        (iterations on a dataset).
        """

        x = x.reshape((1, x.shape[0], 1))
        y = y.reshape((1, y.shape[0], 1))

        y_orig = []
        y_pred = []
        cur_pred = 0.0
        ts_timestamp = []
        outlier_flag = [0] * (x.shape[1] + 1)
        conf_level = 0.005
        win_susp = [0] * self.window_size
        resi_mean = x[:, 0:1, :][0][0][0]
        resi_sqr = x[:, 0:1, :][0][0][0] ** 2
        resi_var = 1.0
        seg_cnt = 1
        pre_susp_bool = False
        sus_point_bool = False

        for i in range(x.shape[1] - 1):
            cur_trnx = x[:, i : i + 1, :]
            cur_trny = y[:, i : i + 1, :]

            tmp_cur_trnx = copy.deepcopy(x[:, i : i + 1, :])
            tmp_cur_trny = copy.deepcopy(y[:, i : i + 1, :])

            weight = 1.0
            sus_point_bool = False

            if pre_susp_bool is True:
                tmp_cur_trnx[0][0][0] = cur_pred

            if i < self.window_size + 1:
                self.train_on_batch(cur_trnx, cur_trny)
                continue
            else:
                cur_prediction = self.predict_on_batch(tmp_cur_trnx)
                cur_pred_ini = cur_prediction[0][0][0]
                tmpresi = tmp_cur_trny[0][0][0] - cur_pred_ini

                (
                    norm_diff,
                    susp_diff,
                    point_diff,
                ) = self.sliding_window_features(i + 1, x, y, win_susp)

                tmpw = self.sliding_window_weight(
                    norm_diff,
                    susp_diff,
                    point_diff,
                )

                if len(norm_diff) == 0:
                    norm_diff = [0.0]

                testval = point_diff * 1.0 / (sum(norm_diff) / len(norm_diff))
                curmean = resi_mean
                curvar = resi_var
                tmp_zval = (tmpresi - curmean) * 1.0 / sqrt(curvar)
                tmp_pro_conve = st.norm.cdf(tmp_zval)

                if tmp_pro_conve > (1 - conf_level) or tmp_pro_conve < conf_level:
                    sus_point_bool = True
                    pre_susp_bool = True
                    weight = tmpw
                    win_susp.append(1)
                    win_susp.pop(0)

                    if outlier_flag[i + 1] == 1:
                        print(
                            "outlier weight at ",
                            i + 1,
                            ": YES ",
                            weight,
                            point_diff,
                            sum(win_susp),
                            sum(norm_diff),
                            testval,
                        )
                    else:
                        pass
                        print(
                            "suspicous points at ",
                            i + 1,
                            ": ",
                            weight,
                            point_diff,
                            sum(win_susp),
                            sum(norm_diff),
                            testval,
                        )
                else:
                    pre_susp_bool = False
                    win_susp.append(0)
                    win_susp.pop(0)
                    if outlier_flag[i + 1] == 1:
                        print(
                            "outlier weight at ",
                            i + 1,
                            ": NO ",
                            weight,
                            point_diff,
                            sum(win_susp),
                            sum(norm_diff),
                            testval,
                        )
                if i in ts_timestamp:
                    print(
                        "----------- CHANGE POINT AT -------",
                        i,
                        ":",
                        tmp_pro_conve,
                        weight,
                    )
                self.train_on_batch(tmp_cur_trnx, tmp_cur_trny * (weight))

            cur_pred = self.predict_on_batch(tmp_cur_trnx)[0][0][0]
            tmpresi = weight * (tmp_cur_trny[0][0][0] - cur_pred)
            resi_mean = resi_mean * seg_cnt * 1.0 / (seg_cnt + 1) + tmpresi / (
                seg_cnt + 1
            )
            resi_sqr = resi_sqr * seg_cnt * 1.0 / (seg_cnt + 1) + tmpresi * tmpresi / (
                seg_cnt + 1
            )
            resi_var = resi_sqr - resi_mean * resi_mean
            seg_cnt = seg_cnt + 1

            # ----------------prediction-------------------------- #
            if sus_point_bool is True:
                vali_testx = copy.deepcopy(x[:, i : i + 1, :])
                vali_testx[0][0][0] = cur_pred_ini
            else:
                vali_testx = copy.deepcopy(x[:, i + 1 : i + 2, :])

            if outlier_flag[i + 2] != 1:
                pred_test = self.predict_on_batch(vali_testx)[0][0][0]
                y_orig.append(y[:, i + 1 : i + 2, :][0][0][0])
                y_pred.append(pred_test)

        print("Normalised MSE: ", nmse(y_orig, y_pred))

        # plot_series(
        #     y_orig,
        #     y_pred,
        #     plot_title="Online Prediction by WGSLTM on Bike rental",
        #     save_path="plot_wglstm.png",
        # )

        return SimpleNamespace(history={m.name: m.result() for m in self.metrics})

    def sliding_window_features(self, cur_pos, dataX, dataY, susp_list):
        winsize = self.window_size
        resi_list = []
        for i in range(cur_pos - winsize, cur_pos):
            cur_trnx = dataX[:, i : i + 1, :]
            cur_trny = dataY[:, i : i + 1, :]
            cur_pred = self.predict_on_batch(cur_trnx)[0][0][0]
            tmpresi = cur_trny[0][0][0] - cur_pred * 1.0
            resi_list.append(tmpresi)

        # extract features
        tmparr = list(zip(range(winsize), susp_list))
        lnormal = -1
        rnormal = -1
        rsus = -1

        tmpdiff = []
        for i in range(cur_pos - winsize, cur_pos):
            cur_val = dataX[:, i : i + 1, :][0][0][0]
            if susp_list[i - (cur_pos - winsize)] == 0:
                if lnormal == -1:
                    lnormal = i
                    tmpdiff.append(0.0)
                else:
                    tmpdiff.append(
                        abs(cur_val - dataX[:, lnormal : lnormal + 1, :][0][0][0])
                    )
                    lnormal = i
            else:
                rsus = i
                if i >= (cur_pos - winsize):
                    tmpdiff.append(abs(cur_val - dataX[:, i - 1 : i, :][0][0][0]))
                else:
                    tmpdiff.append(abs(0.0))

        for i in range(cur_pos - winsize, cur_pos):
            tmp_dta = cur_pos - (i - (cur_pos - winsize)) - 1
            cur_val = dataX[:, tmp_dta : tmp_dta + 1, :][0][0][0]
            tmp_win = tmp_dta - (cur_pos - winsize)
            if susp_list[tmp_win] == 0:
                if rnormal == -1:
                    rnormal = tmp_dta
                else:
                    tmpdiff[tmp_win] = tmpdiff[tmp_win] + (
                        abs(cur_val - dataX[:, rnormal : rnormal + 1, :][0][0][0])
                    )
                    rnormal = tmp_dta
            else:
                if tmp_dta <= (cur_pos - 1):
                    tmpdiff[tmp_win] = tmpdiff[tmp_win] + (
                        abs(cur_val - dataX[:, tmp_dta + 1 : tmp_dta + 2, :][0][0][0])
                    )

        tmparr = list(zip(tmpdiff, susp_list))
        norm_diff = [i[0] for i in tmparr if i[1] == 0]
        susp_diff = [i[0] for i in tmparr if i[1] == 1]

        tmpdiff = -1
        if rsus != -1:
            tmpdiff = abs(
                dataX[:, cur_pos : cur_pos + 1, :][0][0][0]
                - dataX[:, rsus : rsus + 1, :][0][0][0]
            )
        tmpval = abs(
            dataX[:, cur_pos : cur_pos + 1, :][0][0][0]
            - dataX[:, cur_pos - 1 : cur_pos, :][0][0][0]
        )
        tmpdiff = max(tmpdiff, tmpval)

        return norm_diff, susp_diff, tmpdiff

    def sliding_window_weight(self, normDiff, suspDiff, pnt_diff):
        weight_mag = 5
        weight_beta = 0.05

        if len(suspDiff) != 0 and len(normDiff) != 0:
            r2 = 1.0 * sum(suspDiff) * len(normDiff) / sum(normDiff) / len(suspDiff)
            if r2 < 5:
                r2 = 0.0

        if len(normDiff) != 0:
            tmpr = pnt_diff * 1.0 / (sum(normDiff) / len(normDiff))
            if tmpr > weight_mag:
                tmpw = np.exp(-1 * weight_beta * tmpr)
            else:
                tmpw = 1.0
        else:
            tmpw = 1.0

        return tmpw


class WGLSTM(KerasRegressor, KerasSerializationMixin):
    """
    Keras wrapper for Weighted Gradient learning based LSTM (WGLSTM)
    neural network in online learning of time series.

    WGLSTM is an adaptive gradient learning method for recurrent neural
    networks (RNN) to forecast streaming time series in the presence of
    anomalies and change points.

    It leverage local features, which are extracted from a sliding window
    over time series, to dynamically weight the gradient, at each time instant
    for updating the current neural network.

    Parameters
    ----------

        optimizer : Union[str, tf.keras.optimizers.Optimizer,
        Type[tf.keras.optimizers.Optimizer]], default "sgd"
            This can be a string for Keras' built in optimizersan instance of
            tf.keras.optimizers.Optimizer or a class inheriting from
            tf.keras.optimizers.Optimizer. Only strings and classes support
            parameter routing.

        loss : Union[Union[str, tf.keras.losses.Loss, Callable], None],
        default="mse"
            The loss function to use for training. This can be a string for
            Keras' built in losses, an instance of tf.keras.losses.Loss or a
            class inheriting from tf.keras.losses.Loss .

        metrics : List[str], default=None
            List of metrics to evaluate and report at each epoch.

        hidden_layer_activation : str, default=linear
            What activation to use on the hidden layer - can use any
            tf.keras.activation function name.

        hidden_layer_neurons: int, default=450
            number of neurons to use for learning in the hidden layer.

        regularization_factor: float, default=0.0001
            The regularization factor to used during the training.

        timesteps: int, default=1
            The amount of time steps to run your recurrent neural network.

        num_of_features : int, default=1
            Number of feature variables in every time step.

        window_size: int, default=20
            The size of the sliding window for feature extraction.

        epochs: int, default=1
            Number of training steps.

        verbose: int default=0
            0 means no output printed during training.

    """

    def __init__(
        self,
        loss="mse",
        optimizer=tf.keras.optimizers.SGD(
            learning_rate=0.002, momentum=0.03, nesterov=True
        ),
        metrics=None,
        epochs=1,
        verbose=0,
        num_of_features=1,
        hidden_layer_neurons=450,
        hidden_layer_activation="linear",
        regularization_factor=0.0001,
        timesteps=1,
        window_size=20,
        **kwargs,
    ) -> None:
        super(WGLSTM, self).__init__(
            _Model(
                num_of_features,
                hidden_layer_neurons,
                hidden_layer_activation,
                regularization_factor,
                timesteps,
                window_size,
            ),
            loss=loss,
            optimizer=optimizer,
            metrics=metrics,
            epochs=epochs,
            verbose=verbose,
            **kwargs,
        )



from river.compat import River2SKLClassifier, River2SKLRegressor

from sail.models.base import SAILModel
from sail.utils.logging import configure_logger
from sail.common.mixin import RiverAttributeMixin

LOGGER = configure_logger(logger_name="River")


class SailRiverRegressor(RiverAttributeMixin, River2SKLRegressor, SAILModel):
    def __init__(self, *args, **Kwargs):
        super(SailRiverRegressor, self).__init__(*args, **Kwargs)


class SailRiverClassifier(RiverAttributeMixin, River2SKLClassifier, SAILModel):
    def __init__(self, *args, **Kwargs):
        super(SailRiverClassifier, self).__init__(*args, **Kwargs)



import torch
import torch.nn as nn
from skorch.regressor import NeuralNetRegressor
import torch.nn.functional as F
from sail.models.torch.base import TorchSerializationMixin


class NNModel(nn.Module):
    def __init__(
        self,
        num_units=10,
        nonlin=F.relu,
    ):
        super(NNModel, self).__init__()
        self.num_units = num_units
        self.nonlin = nonlin

        self.dense0 = nn.Linear(20, num_units)
        self.nonlin = nonlin
        self.dense1 = nn.Linear(num_units, 10)
        self.output = nn.Linear(10, 1)

    def forward(self, X, **kwargs):
        X = self.nonlin(self.dense0(X))
        X = F.relu(self.dense1(X))
        X = self.output(X)
        return X


class NNRegressor(NeuralNetRegressor, TorchSerializationMixin):
    def __init__(self, ni, no, nh, nlayers, module=NNModel, **kwargs):
        super(NNRegressor, self).__init__(
            module=module,
            module__ni=ni,
            module__no=no,
            module__nh=nh,
            module__nlayers=nlayers,
            **kwargs
        )


if __name__ == "__main__":
    regressor = NNRegressor(None, None, None, None)



from .base import SAILTransformer
import pandas as pd
from sklearn.base import ClassNamePrefixFeaturesOutMixin
from sklearn.utils.validation import check_is_fitted


class EncodeDateTransformer(ClassNamePrefixFeaturesOutMixin, SAILTransformer):
    def __init__(self, datetime_col, temporal_fields, prefix=None) -> None:
        super(EncodeDateTransformer, self).__init__()
        self.datetime_col = datetime_col
        self.temporal_fields = temporal_fields
        self.prefix = prefix
        self.col = (prefix if prefix else datetime_col) + "_"
        self.validation_params["cast_to_ndarray"] = False

    def _partial_fit(self, X, y=None):
        return self

    def _transform(self, X, y=None, copy=None):
        if self.datetime_col not in X.columns:
            raise Exception(f"Datetime column: {self.datetime_col} not available in X.")

        new_df = pd.DataFrame()
        for field in self.temporal_fields:
            if "time" == field:
                new_df[self.col + field] = (
                    X[self.datetime_col].dt.hour + X[self.datetime_col].dt.minute / 60.0
                )

            if "day" == field:
                new_df[self.col + field] = X[self.datetime_col].dt.dayofweek

            if "month" == field:
                new_df[self.col + field] = X[self.datetime_col].dt.month - 1

            if "is_weekend" == field:
                new_df[self.col + field] = (
                    X[self.datetime_col].dt.dayofweek >= 5
                ).astype("int")

        self.features_ = list(new_df.columns)
        return new_df.to_numpy()

    def get_feature_names_out(self, input_features=None):
        """Get output feature names for transformation.

        Parameters
        ----------
        input_features : array-like of str or None, default=None
            Input features.

            - If `input_features` is `None`, then `feature_names_in_` is
              used as feature names in. If `feature_names_in_` is not defined,
              then the following input feature names are generated:
              `["x0", "x1", ..., "x(n_features_in_ - 1)"]`.
            - If `input_features` is an array-like, then `input_features` must
              match `feature_names_in_` if `feature_names_in_` is defined.

        Returns
        -------
        feature_names_out : ndarray of str objects
            Same as input features.
        """
        check_is_fitted(self, "n_features_in_")
        return self.features_



import glob
import os
import time

import numpy as np
import torch
from torch.utils.tensorboard import FileWriter, SummaryWriter
from torch.utils.tensorboard.summary import scalar

from sail.utils.logging import configure_logger

LOGGER = configure_logger(logger_name="TensorboardWriter")


class TensorboardWriter(SummaryWriter):
    def __init__(
        self, log_dir=None, exp_name="Training_Logs", use_dir=None, *args, **kwrags
    ):
        if use_dir:
            exp_dir = log_dir + "/" + use_dir
        else:
            dirs = sorted(
                glob.glob(os.path.join(log_dir, exp_name, "v*")), reverse=True
            )
            new_dir_no = 1
            if len(dirs) > 0:
                new_dir_no = int(dirs[0].split("/")[-1][1]) + 1
            exp_dir = os.path.join(log_dir, exp_name, "v" + str(new_dir_no))
            # log_dir + "/" + exp_name + "/_v" + str(new_dir_no)

        super().__init__(exp_dir, *args, **kwrags)
        self.all_writers = {}

        LOGGER.info(
            f"Sending training output to Tensorboard logs. Please run `tensorboard --logdir {os.path.join(log_dir, exp_name)}` in terminal to start tensorboard server and track training progress."
        )

    def get_state(self):
        return {"log_dir": self.log_dir}

    def set_state(self, state):
        self.log_dir = state["log_dir"]

    def _get_file_writer(self):
        ...

    def write_predictions(self, y_pred, y_true, start_index):
        for index, (yh, y_true) in enumerate(zip(y_pred, y_true), start=start_index):
            self.add_scalars_custom(
                "Predictions",
                {
                    "y_true": yh,
                    "y_pred": y_true,
                },
                index,
            )

    def write_score(self, score, epoch_n, drift_point=False):
        self.add_scalar(
            "Score_and_Detect",
            score,
            epoch_n,
        )

        if drift_point:
            self.add_scalar(
                "Score_and_Detect",
                np.nan,
                epoch_n,
            )
            self.add_scalar(
                "Score_and_Detect",
                score,
                epoch_n,
            )

        #     self.add_scalars_custom(
        #         "Score_and_Detect",
        #         {
        #             "Drift": np.nan,
        #         },
        #         epoch_n,
        #         include_main_tag=False,
        #     )
        #     self.add_scalars_custom(
        #         "Score_and_Detect",
        #         {
        #             "Drift": score,
        #         },
        #         epoch_n,
        #         include_main_tag=False,
        #     )
        self.flush()

    def add_scalar(
        self,
        tag,
        scalar_value,
        global_step=None,
        walltime=None,
        new_style=False,
        double_precision=False,
    ):
        """Add scalar data to summary.

        Args:
            tag (str): Data identifier
            scalar_value (float or string/blobname): Value to save
            global_step (int): Global step value to record
            walltime (float): Optional override default walltime (time.time())
            with seconds after epoch of event
            new_style (boolean): Whether to use new style (tensor field) or old
            style (simple_value field). New style could lead to faster data loading.
        Examples::

            from torch.utils.tensorboard import SummaryWriter
            writer = SummaryWriter()
            x = range(100)
            for i in x:
                writer.add_scalar('y=2x', i * 2, i)
            writer.close()

        Expected result:

        .. image:: _static/img/tensorboard/add_scalar.png
        :scale: 50 %

        """
        torch._C._log_api_usage_once("tensorboard.logging.add_scalar")
        summary = scalar(
            tag, scalar_value, new_style=new_style, double_precision=double_precision
        )
        fw_tag = self.log_dir + "/" + tag
        if self.all_writers and fw_tag in self.all_writers.keys():
            fw = self.all_writers[fw_tag]
        else:
            fw = FileWriter(
                fw_tag, self.max_queue, self.flush_secs, self.filename_suffix
            )
            self.all_writers[fw_tag] = fw

        fw.add_summary(summary, global_step, walltime)

    def add_scalars_custom(
        self,
        main_tag,
        tag_scalar_dict,
        global_step=None,
        walltime=None,
        include_main_tag=True,
    ):
        """Adds many scalar data to summary.

        Args:
            main_tag (str): The parent name for the tags
            tag_scalar_dict (dict): Key-value pair storing the tag and corresponding values
            global_step (int): Global step value to record
            walltime (float): Optional override default walltime (time.time())
              seconds after epoch of event

        Examples::

            from torch.utils.tensorboard import SummaryWriter
            writer = SummaryWriter()
            r = 5
            for i in range(100):
                writer.add_scalars('run_14h', {'xsinx':i*np.sin(i/r),
                                                'xcosx':i*np.cos(i/r),
                                                'tanx': np.tan(i/r)}, i)
            writer.close()
            # This call adds three values to the same scalar plot with the tag
            # 'run_14h' in TensorBoard's scalar section.

        Expected result:

        .. image:: _static/img/tensorboard/add_scalars.png
           :scale: 50 %

        """
        torch._C._log_api_usage_once("tensorboard.logging.add_scalars")
        walltime = time.time() if walltime is None else walltime
        for tag, scalar_value in tag_scalar_dict.items():
            if include_main_tag:
                fw_tag = self.log_dir + "/" + main_tag + "_" + tag
            else:
                fw_tag = self.log_dir + "/" + tag
            if self.all_writers and fw_tag in self.all_writers.keys():
                fw = self.all_writers[fw_tag]
            else:
                fw = FileWriter(
                    fw_tag, self.max_queue, self.flush_secs, self.filename_suffix
                )
                self.all_writers[fw_tag] = fw

            fw.add_summary(scalar(main_tag, scalar_value), global_step, walltime)



# Implementation of Time Series Classification from Scratch with Deep Neural Networks: A Strong Baseline (2016, arXiv) in PyTorch.
# The scripts are provided by https://github.com/aybchan/time-series-classification
import os
import subprocess

import numpy as np
import torch
import torch.nn.functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm

from sail.models.torch.data import Data


def download_datasets(datasets):
    """
    Download the time series datasets used in the paper
    from http://www.timeseriesclassification.com
    NonInvThorax1 and NonInvThorax2 are missing
    """
    subprocess.call("mkdir -p data".split())
    datasets_pbar = tqdm(datasets)
    print(datasets_pbar)
    for dataset in datasets_pbar:
        datasets_pbar.set_description("Downloading {0}".format(dataset))
        subprocess.call(
            "curl http://www.timeseriesclassification.com/Downloads/{0}.zip -o data/{0}.zip".format(
                dataset
            ).split()
        )
        datasets_pbar.set_description("Extracting {0} data".format(dataset))
        subprocess.call("unzip data/{0} -d data/".format(dataset).split())
        assert os.path.exists("data/{}_TRAIN.arff".format(dataset)), dataset


def data_dictionary(dataset_path, datasets):
    """
    Create a dictionary of train/test DataLoaders for
    each of the datasets downloaded
    """
    dataset_dict = {}
    pbar = tqdm(datasets)
    print(pbar)
    for dataset in pbar:
        pbar.set_description("Processing {}".format(dataset))
        train_set, test_set = Data(
            os.path.join(dataset_path, dataset), False
        ), Data(os.path.join(dataset_path, dataset), True)
        batch_size = min(16, len(train_set) // 10)

        dataset_dict[dataset] = {}
        dataset_dict[dataset]["train"] = DataLoader(
            train_set, batch_size=batch_size
        )
        dataset_dict[dataset]["test"] = DataLoader(
            test_set, batch_size=batch_size
        )

    return dataset_dict



# Implementation of Time Series Classification from Scratch with Deep Neural Networks: A Strong Baseline (2016, arXiv) in PyTorch.
# The scripts are provided by https://github.com/aybchan/time-series-classification
import os

import numpy as np
import pandas as pd
import torch
from scipy.io.arff import loadarff
from torch.utils.data import Dataset


class Data(Dataset):
    """
    Implements a PyTorch Dataset class
    """

    def __init__(self, dataset, testing):
        train_data = pd.DataFrame(loadarff(dataset + "_TRAIN.arff")[0])
        dtypes = {i: np.float32 for i in train_data.columns[:-1]}
        dtypes.update({train_data.columns[-1]: np.int})
        train_data = train_data.astype(dtypes)

        self.x = train_data.iloc[:, :-1].values
        self.y = train_data.iloc[:, -1].values

        self.mean = np.mean(self.x, axis=0)
        self.std = np.std(self.x, axis=0)
        if testing:
            test_data = pd.DataFrame(loadarff(dataset + "_TEST.arff")[0])
            dtypes = {i: np.float32 for i in test_data.columns[:-1]}
            dtypes.update({test_data.columns[-1]: np.int})
            test_data = test_data.astype(dtypes)

            self.x = test_data.iloc[:, :-1].values
            self.y = test_data.iloc[:, -1].values

        class_labels = np.unique(self.y)
        if (class_labels != np.arange(len(class_labels))).any():
            new_labels = {
                old_label: i for i, old_label in enumerate(class_labels)
            }
            self.y = [new_labels[old_label] for old_label in self.y]

        self.x = (self.x - self.mean) / self.std

    def __len__(self):
        return len(self.y)

    def __getitem__(self, idx):
        x = torch.Tensor([self.x[idx]]).float()
        y = torch.Tensor([self.y[idx]]).long()
        return x, y



import os
from sail.utils.serialization import save_obj, load_obj
from sail.utils.logging import configure_logger

LOGGER = configure_logger(logger_name="Sklearn")


def save(model, model_folder):
    """
    Saves the model to joblib format.

    Args:
        model_folder: String, PathLike, path to SavedModel.
    """
    metadata_path = os.path.join(model_folder, "metadata")
    save_obj(model, metadata_path, "model", serialize_type="joblib")
    LOGGER.info("Model saved successfully.")


def load(model_folder):
    """
    Load the Model from the mentioned folder location

    Args:
        model_folder: One of the following:
            - String or `pathlib.Path` object, path to the saved model
    """
    metadata_path = os.path.join(model_folder, "metadata")
    model = load_obj(metadata_path, "model", serialize_type="joblib")
    LOGGER.info("Model loaded successfully.")
    return model



# PyTorch implementation for TCN
# This implementation is based on https://github.com/unit8co/darts/blob/master/darts/models/forecasting/tcn_model.py
# Paper https://arxiv.org/abs/1803.01271
# Paper repository: https://github.com/locuslab/TCN
# By Kasper Hjort Berthelsen and Mads Ehrhorn

from skorch.regressor import NeuralNetRegressor

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.weight_norm import weight_norm as wn


class _ResidualBlock(nn.Module):
    def __init__(
        self,
        num_filters: int,
        kernel_size: int,
        dilation_base: int,
        dropout_fn: nn.Dropout,
        weight_norm: bool,
        nr_blocks_below: int,
        num_layers: int,
        input_size: int,
        target_size: int,
    ):
        """PyTorch module implementing a residual block module used in `TCNModel`.
        This implementation is based on https://github.com/unit8co/darts/blob/master/darts/models/forecasting/tcn_model.py

        Args:
        num_filters (int): The number of filters in a convolutional layer of the TCN.
        kernel_size (int): The size of every kernel in a convolutional layer.
        dilation_base (int): The base of the exponent that will determine the dilation on every level.
        dropout_fn (nn.Dropout): The dropout function to be applied to every convolutional layer.
        weight_norm (bool): Boolean value indicating whether to use weight normalization.
        nr_blocks_below (int): The number of residual blocks before the current one.
        num_layers (int): The number of convolutional layers.
        input_size (int): The dimensionality of the input time series of the whole network.
        target_size (int): The dimensionality of the output time series of the whole network.
        """
        super().__init__()

        self.dilation_base = dilation_base
        self.kernel_size = kernel_size
        self.dropout_fn = dropout_fn
        self.num_layers = num_layers
        self.nr_blocks_below = nr_blocks_below

        input_dim = input_size if nr_blocks_below == 0 else num_filters
        output_dim = target_size if nr_blocks_below == num_layers - 1 else num_filters
        self.conv1 = nn.Conv1d(
            input_dim,
            num_filters,
            kernel_size,
            dilation=(dilation_base ** nr_blocks_below),
        )
        self.conv2 = nn.Conv1d(
            num_filters,
            output_dim,
            kernel_size,
            dilation=(dilation_base ** nr_blocks_below),
        )
        if weight_norm:
            self.conv1, self.conv2 = wn(self.conv1), wn(self.conv2)

        if input_dim != output_dim:
            self.conv3 = nn.Conv1d(input_dim, output_dim, 1)

    def forward(self, x):
        """_ResidualBlock forward pass.

        Args:
            x (torch.Tensor): The input data

        Shape:
            - Input: `(batch_size, in_dimension, input_chunk_length)`
            - Output: `(batch_size, out_dimension, input_chunk_length)`
        """
        residual = x

        left_padding = (self.dilation_base ** self.nr_blocks_below) * (
            self.kernel_size - 1
        )
        x = F.pad(x, (left_padding, 0))
        x = self.dropout_fn(F.relu(self.conv1(x)))

        x = F.pad(x, (left_padding, 0))
        x = self.conv2(x)
        if self.nr_blocks_below < self.num_layers - 1:
            x = F.relu(x)
        x = self.dropout_fn(x)

        if self.conv1.in_channels != self.conv2.out_channels:
            residual = self.conv3(residual)
        x = x + residual

        return x


class TCNModel(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        kernel_size: int,
        num_filters: int,
        num_layers: int,
        dilation_base: int,
        weight_norm: bool,
        dropout: float,
    ):
        """PyTorch module implementing a dilated TCN module used in `TCNModel`.
        This implementation is based on https://github.com/unit8co/darts/blob/master/darts/models/forecasting/tcn_model.py

        Args:
        input_dim (int): The dimensionality of the input time series.
        output_dim (int): The dimensionality of the output time series.
        kernel_size (int): The size of every kernel in a convolutional layer.
        num_filters (int): The number of filters in a convolutional layer of the TCN.
        num_layers (int): The number of convolutional layers.
        dilation_base (int): The base of the exponent that will determine the dilation on every level.
        weight_norm (bool): Boolean value indicating whether to use weight normalization.
        dropout (float): The dropout rate for every convolutional layer.
        """

        super().__init__()

        # Defining parameters
        self.input_size = input_dim
        self.n_filters = num_filters
        self.kernel_size = kernel_size
        self.target_size = output_dim
        self.dilation_base = dilation_base
        self.dropout = nn.Dropout(p=dropout)
        self.num_layers = num_layers

        # Building TCN module
        self.res_blocks_list = []
        for i in range(num_layers):
            res_block = _ResidualBlock(
                num_filters,
                kernel_size,
                dilation_base,
                self.dropout,
                weight_norm,
                i,
                num_layers,
                self.input_size,
                self.target_size,
            )
            self.res_blocks_list.append(res_block)
        self.res_blocks = nn.ModuleList(self.res_blocks_list)

    def forward(self, x):
        """TCNModel forward pass.

        Args:
            x (torch.Tensor): The input data

        Shape:
            - Input: `(batch_size, in_dimension, input_chunk_length)`
            - Output: `(batch_size, out_dimension, input_chunk_length)`
        """
        # To aid inputs that omit chunk length, we add a dummy dimension if ndim = 2
        if x.dim() == 2:
            x = torch.unsqueeze(x, dim=-1)
        batch_size = x.size(0)

        for res_block in self.res_blocks_list:
            x = res_block(x)

        x = x.view(batch_size, self.target_size)

        return x


class TCNRegressor(NeuralNetRegressor):
    """Basic NeuralNetRegressor wrapper for TCNModel.

    Args:
        input_dim (int): The dimensionality of the input time series.
        output_dim (int): The dimensionality of the output time series.
        kernel_size (int): The size of every kernel in a convolutional layer.
        num_filters (int): The number of filters in a convolutional layer of the TCN.
        num_layers (int): The number of convolutional layers.
        dilation_base (int): The base of the exponent that will determine the dilation on every level.
        weight_norm (bool): Boolean value indicating whether to use weight normalization.
        dropout (float): The dropout rate for every convolutional layer.
        **kwargs: Arbitrary keyword arguments
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        kernel_size: int = 3,
        num_filters: int = 3,
        num_layers: int = 3,
        dilation_base: int = 2,
        weight_norm: bool = False,
        dropout: float = 0.2,
        **kwargs
    ):
        super(TCNRegressor, self).__init__(
            module=TCNModel,
            module__input_dim=input_dim,
            module__output_dim=output_dim,
            module__kernel_size=kernel_size,
            module__num_filters=num_filters,
            module__num_layers=num_layers,
            module__dilation_base=dilation_base,
            module__weight_norm=weight_norm,
            module__dropout=dropout,
            train_split=None,
            max_epochs=1,
            batch_size=20,
            **kwargs
        )



from numpy.lib.scimath import sqrt


def nmse(y, yh):
    """MSE normalized by the total count."""

    count = len(y)
    nmse = sqrt(
        sum([(y[i] - yh[i]) * (y[i] - yh[i]) for i in range(count)]) * 1.0 / count
    )
    return nmse



# Implementation of Time Series Classification from Scratch with Deep Neural Networks: A Strong Baseline (2016, arXiv) in PyTorch.
# The model is provided by https://github.com/aybchan/time-series-classification 
# The model has been modified slightly changing the stride to 1 by Hao Miao
import torch
import torch.nn as nn
import torch.nn.functional as F
from skorch import NeuralNetClassifier

class _ConvNetModel(nn.Module):
    """ Basic ConvVet: A strong Baseline for Time Series Classification model.

    Args:
        n_in (int): Number of input units
        n_classes (int): Number of classes
    """    
    def __init__(self, n_in, n_classes):
        super(_ConvNetModel,self).__init__()
        self.n_in = n_in
        self.n_classes = n_classes

        self.conv1 = nn.Conv2d(  1,128,(7,1),1,(3,0))
        self.bn1   = nn.BatchNorm2d(128)

        self.conv2 = nn.Conv2d(128,256,(5,1),1,(2,0))
        self.bn2   = nn.BatchNorm2d(256)

        self.conv3 = nn.Conv2d(256,128,(3,1),1,(1,0))
        self.bn3   = nn.BatchNorm2d(128)

        self.fc4   = nn.Linear(128,self.n_classes)


    def forward(self, x: torch.Tensor):
        x = x.view(-1,1,self.n_in,1)

        x = F.relu(self.bn1(self.conv1(x)))
        x = F.relu(self.bn2(self.conv2(x)))
        x = F.relu(self.bn3(self.conv3(x)))

        x = F.avg_pool2d(x,1) #modified
        x = torch.mean(x,dim=2)
        x = x.view(-1,128)
        x = self.fc4(x)
        print()
        return F.log_softmax(x,1)
    

class ConvNet(NeuralNetClassifier):
    """Basic TSC model.
    Args:
        n_in (int): Number of input units
        n_classes (int): Number of classes
    """
    def __init__(self, n_in, n_classes):
        super(ConvNet, self).__init__(
            module=_ConvNetModel,
            module__n_in= n_in,
            module__n_classes=n_classes,
            max_epochs=10, 
            lr=0.01, 
            batch_size=12, 
            optimizer=torch.optim.Adam,
            criterion=torch.nn.CrossEntropyLoss,
            # Shuffle training data on each epoch
            iterator_train__shuffle=False )






import math
import typing
from river import base
from river import optim
from .base import BaseAggregator
from river.compat import convert_river_to_sklearn
import ray
import numpy as np


@ray.remote
def _incremental_learn(estimator, X, length, y_pred_mean, weight, y, learning_rate):
    # X = ray.get(X_id)
    y_pred = estimator.predict(X)
    y_pred_mean += weight * (y_pred - y_pred_mean) / length
    loss = np.average((y - y_pred) ** 2, axis=0)  # self.loss(y_true=y, y_pred=y_pred)
    weight *= math.exp(-learning_rate * loss)
    estimator.partial_fit(X, y)
    return [estimator, weight, y_pred_mean]


__all__ = ["DistEWARegressor"]


class DistEWARegressor(BaseAggregator):
    def __init__(
        self,
        estimators,
        loss= None,
        learning_rate=0.5,
    ):
        """
        :param estimators: Estimator objects with partial_fit defined
                    (incremental learning models).
        :param loss: loss function to be minimized
        :param learning_rate: The learning rate by which the model weights are multiplied at each iteration.
        """

        # if len(estimators) < 2:
        #     raise NotEnoughModels(n_expected=2, n_obtained=len(estimators))

        self.loss = optim.losses.Squared() if loss is None else loss
        self.learning_rate = learning_rate
        self.weights = [1.0] * len(estimators)

        estimators = [convert_river_to_sklearn(est) if "river" in est.__module__ else est
                      for est in estimators]
        # self.estimators = [est.remote() for est in estimators]
        super().__init__(estimators)

    def _partial_fit(self, X, y=None, **kwargs):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param kwargs:
        :return: numpy.ndarray of shape (n_samples)
        """
        y_pred_mean = 0.0

        # Make a prediction and update the weights accordingly for each model
        length = len(self)
        total = 0
        obj_ids = []

        for i, estimator in enumerate(self.base_estimators):
            estimator.n_features_in_ = X.shape[1]
            X_id = ray.put(X)
            obj_ids.append(_incremental_learn.remote(estimator, X_id,
                                                     length, y_pred_mean, self.weights[i],
                                  y, self.learning_rate))
        ray_values = ray.get(obj_ids)
        self.base_estimators = [i[0] for i in ray_values]
        self.weights = [i[1] for i in ray_values]
        y_pred_mean = [i[2] for i in ray_values]
        y_pred_mean = np.sum(y_pred_mean)
        total = sum(self.weights)
        # Normalize the weights so that they sum up to 1
        if total:
            for i, _ in enumerate(self.weights):
                self.weights[i] /= total

        return y_pred_mean

    def partial_fit(self, X, y=None, classes=None):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :param y: numpy.ndarray of shape (n_samples)
               Labels for the target variable.
        :param kwargs:
        :return: numpy.ndarray of shape (n_samples)
        """

        self._partial_fit(X, y)
        return self

    def predict(self, X):
        """
        :param X: numpy.ndarray of shape (n_samples, n_features).
               Input samples.
        :return: numpy array of shape (n_samples,).
             Class labels/predictions for input samples.
        """
        for i, estimator in enumerate(self.base_estimators):
            if not hasattr(estimator, 'n_features_in_'):
                estimator.n_features_in_ = X.shape[1]
        return sum(
            model.predict(X) * weight for model, weight in zip(self.base_estimators, self.weights)
        )


from tqdm import tqdm

BAR_FORMAT = {
    "pipeline_training": "{desc}: {percentage:3.0f}%{bar} [Steps: {n_fmt}/{total_fmt}, ETA: {elapsed}<{remaining}, Elapsed:{elapsed_s:3.3f}s{postfix}]",
    "model_training": "{desc}: {percentage:3.0f}%{bar} [ETA: {elapsed}<{remaining}, Elapsed:{elapsed_s:3.3f}s{postfix}]",
    "scoring": "{desc}: {percentage:3.0f}%{bar} [Points: {n_fmt}/{total_fmt}, Elapsed:{elapsed_s:3.4f}s{postfix}]",
    "progressive_score": "{desc} [Elapsed: {elapsed_s:3.5f}s{postfix}]",
}


class SAILProgressBar:
    def __init__(self, steps, desc, params, format, verbose=1) -> None:
        self.remaining_steps = steps
        self.desc = desc
        self.params = params
        self.verbose = verbose

        if self.check_verbosity():
            self.pbar = tqdm(
                total=steps,
                ascii=" =",
                bar_format=BAR_FORMAT[format],
            )
            self.pbar.set_description_str(desc, refresh=True)
            self.pbar.set_postfix(params, refresh=True)

    def check_verbosity(self):
        return self.verbose == 1

    def append_desc(self, desc, divider=" "):
        if self.check_verbosity():
            self.pbar.set_description_str(self.desc + divider + desc, refresh=True)

    def update(self, value=1):
        if self.check_verbosity():
            self.remaining_steps -= value
            self.pbar.update(value)

    def update_params(self, key, value):
        if self.check_verbosity():
            self.params[key] = value
            self.pbar.set_postfix(self.params, refresh=True)

    def finish(self):
        if self.check_verbosity():
            self.pbar.update(self.remaining_steps)

    def refresh(self):
        self.pbar.refresh()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        if self.check_verbosity():
            self.append_desc(desc="", divider="")
            return self.pbar.close()

    def close(self):
        if self.check_verbosity():
            self.pbar.close()



import numpy as np
from numpy.linalg import pinv
import math


def forgettingFunction(num_of_samples):
    return math.exp(0.7 * num_of_samples)


def sigmoidActivation(x):
    return 1.0 / (1.0 + math.exp(-x))


def inv_sigmoid(x):
    y = 1.0 / (1.0 + math.exp(-x))
    return math.log((y / (1 - y)))


def derv_sigmoid(x):
    return 1.0 * math.exp(-x) / (1.0 + math.exp(-x)) ** 2


class RoIML_Model:
    def __init__(self, num_x_features, num_output):
        """
        :param x_features: int. (x_i) no. of features in input x also equal to no. of input nodes.
        :param num_output: int. (d_J) or no. of output neurons/nodes
        """
        self.num_xi = num_x_features
        self.num_dj = num_output

        self.A = np.zeros((num_x_features, num_x_features))
        self.b = np.zeros((num_x_features, num_output))
        self.w_ji = np.zeros((num_x_features, num_output))

    def calculate_A_pi(self, s, x_ps, x_is, d_js):
        """
        :param s: sth training sample from the input data
        :param i: number of input features for a given training sample
        :param X: numpy array of shape (num_samples, num_x_features) The input samples.
        :param d: numpy.ndarray of shape (num_samples, j)
               Labels for the target variable/class.
        """

        h_js = forgettingFunction(s)
        # d_js = d[s][j]
        # x_ps = X[s][p]
        dhat_js = inv_sigmoid(d_js)
        df_dhat_js = derv_sigmoid(dhat_js)
        # x_is = X[s][i]
        A_pi = h_js * x_is * x_ps * df_dhat_js * df_dhat_js

        return A_pi

    def calculate_b_pj(self, s, x_ps, d_js):
        """
        :param s: sth training sample from the input data
        :param j: number of output neurons which is the number of classes for classification problems or J=1 for regression problems
        :param X: numpy array of shape (num_samples, num_x_features) The input samples.
        :param d: numpy.ndarray of shape (num_samples, j)
               Labels for the target variable/class.
        """

        # x_ps = X[s][p]
        # d_js = d[s][j]
        h_js = forgettingFunction(s)
        dhat_js = inv_sigmoid(d_js)
        df_dhat_js = derv_sigmoid(dhat_js)
        b_pj = h_js * dhat_js * x_ps * df_dhat_js * df_dhat_js

        return b_pj

    def partial_fit(self, s, X, d_js, j):
        for p in range(self.num_xi):
            x_ps = X[s][p]
            for i in range(self.num_xi):
                x_is = X[s][i]
                A_pis = self.calculate_A_pi(s, x_ps, x_is, d_js)
                self.A[p][i] = self.A[p][i] + A_pis
            b_pjs = self.calculate_b_pj(s, x_ps, d_js)
            self.b[p][j] = self.b[p][j] + b_pjs
        # calculate w_ji using A_pi and b_pj
        inv_A = np.linalg.pinv(self.A)
        self.w_ji = np.dot(inv_A, self.b)
        return self

    def predict(self, test_s):
        return np.dot(test_s, self.w_ji)
